#!/usr/bin/env python
#
# Copyright (c) 2017 10X Genomics, Inc. All rights reserved.
#
from __future__ import absolute_import

import math

PIPELINE_AGGR = "aggr"
PIPELINE_COUNT = "count"
PIPELINE_REANALYZE = "reanalyze"
PIPELINE_VDJ = "vdj"
PIPELINE_META_COUNT = "meta-count"
CELLRANGER_COMMAND_NAME = "Cell Ranger"

# TODO delete this if it isn't being used anywhere
WEBSHIM_PIPELINES = [
    PIPELINE_AGGR,
    PIPELINE_COUNT,
    PIPELINE_REANALYZE,
    PIPELINE_VDJ,
    PIPELINE_META_COUNT,
]

ALARM_WARN = "warn"
ALARM_ERROR = "error"
DEFAULT_CHART_WIDTH = 600
DEFAULT_CHART_HEIGHT = 400
ALARMS = "alarms"

# Formatting of xy coordinates in javascript lists going to plotly
DATA_VALUE_FORMAT = "%.4f"

# these are the options for making the views fixed;
# add these to the modeBarButtonsToRemove list to
# disable movement & zoom
#
# src: src/components/modebar/buttons.js in Plotly
CHARTS_PLOTLY_MODEBAR_TRANSFORM_BUTTONS = [
    "zoom2d",
    "pan2d",
    "zoomIn2d",
    "zoomOut2d",
    "autoScale2d",
    # 'resetScale2d'  can't totally disable interaction, it seems-- keep reset option
]

CHARTS_PLOTLY_EXPORT_BUTTONS = [
    "toImage",
    "sendDataToCloud",
]

CHARTS_PLOTLY_FIXED_CONFIG = {
    "modeBarButtonsToRemove": CHARTS_PLOTLY_MODEBAR_TRANSFORM_BUTTONS
    + CHARTS_PLOTLY_EXPORT_BUTTONS,
    "displaylogo": False,
    "showLink": False,
}

CHARTS_PLOTLY_MOVABLE_CONFIG = {
    "modeBarButtonsToRemove": CHARTS_PLOTLY_EXPORT_BUTTONS,
    "displaylogo": False,
    "showLink": False,
}

BC_RANK_PLOT_LINE_WIDTH = 3
# Gradient scheme used in the barcode rank plot
BC_PLOT_COLORS = [
    "#dddddd",
    "#d1d8dc",
    "#c6d3dc",
    "#bacfdb",
    "#aecada",
    "#a3c5d9",
    "#97c0d9",
    "#8cbbd8",
    "#80b7d7",
    "#74b2d7",
    "#6aadd6",
    "#66abd4",
    "#62a8d2",
    "#5ea5d1",
    "#59a2cf",
    "#559fce",
    "#519ccc",
    "#4d99ca",
    "#4997c9",
    "#4594c7",
    "#4191c5",
    "#3d8dc4",
    "#3a8ac2",
    "#3787c0",
    "#3383be",
    "#3080bd",
    "#2c7cbb",
    "#2979b9",
    "#2676b7",
    "#2272b6",
    "#1f6eb3",
    "#1d6ab0",
    "#1a65ac",
    "#1861a9",
    "#155ca6",
    "#1358a2",
    "#10539f",
    "#0e4f9b",
    "#0b4a98",
    "#094695",
    "#09438f",
    "#0a4189",
    "#0c3f83",
    "#0d3d7c",
    "#0e3b76",
    "#103970",
    "#11366a",
    "#123463",
    "#14325d",
    "#153057",
]


def BC_PLOT_CMAP(density):
    """
    Colormap utility fn to map a number to one of the colors in the gradient
    color scheme defined above
    Input
    - density : A real number in the range [0,1]
    """
    assert density >= 0.0
    assert density <= 1.0
    levels = len(BC_PLOT_COLORS)
    ind = min(levels - 1, int(math.floor(levels * density)))
    return BC_PLOT_COLORS[ind]


HISTOGRAM_METRIC_ORDER_INTEGER_BIN = "order_integer"

# Freq expects ints, prop expects floats
HISTOGRAM_METRIC_ORDER_DECREASING_FREQUENCY = "order_frequency"
HISTOGRAM_METRIC_ORDER_DECREASING_PROPORTION = "order_proportion"

HISTOGRAM_METRIC_DEFAULT_ORDERING = HISTOGRAM_METRIC_ORDER_INTEGER_BIN

HISTOGRAM_METRIC_ORDER = [
    HISTOGRAM_METRIC_ORDER_INTEGER_BIN,
    HISTOGRAM_METRIC_ORDER_DECREASING_FREQUENCY,
    HISTOGRAM_METRIC_ORDER_DECREASING_PROPORTION,
]



#!/usr/bin/env python
#
# Copyright (c) 2017 10X Genomics, Inc. All rights reserved.
#
from __future__ import absolute_import

import collections
import numpy as np
import os
import sys
import bmkcc.cellranger.analysis.io as analysis_io
import bmkcc.cellranger.io as cr_io

CLUSTER_TYPE_KMEANS = "kmeans"
CLUSTER_TYPE_GRAPHCLUST = "graphclust"
CLUSTER_TYPE_KMEDOIDS = "kmedoids"
CLUSTER_TYPE_CELLTYPES = "celltype"

CLUSTERING = collections.namedtuple(
    "CLUSTERING",
    [
        "clusters",
        "num_clusters",
        "cluster_score",
        "clustering_type",
        "global_sort_key",
        "description",
    ],
)


def format_legacy_clustering_key(cluster_type, cluster_param):
    assert cluster_type == CLUSTER_TYPE_KMEANS
    return "_%d" % cluster_param


def format_clustering_key(cluster_type, cluster_param):
    """ Generate a machine-readable string that describes a particular clustering """
    if cluster_type == CLUSTER_TYPE_KMEANS:
        return "%s_%d_clusters" % (CLUSTER_TYPE_KMEANS, cluster_param)
    elif cluster_type == CLUSTER_TYPE_KMEDOIDS:
        return "%s_%d_clusters" % (CLUSTER_TYPE_KMEDOIDS, cluster_param)
    elif cluster_type == CLUSTER_TYPE_GRAPHCLUST:
        return CLUSTER_TYPE_GRAPHCLUST
    elif cluster_type == CLUSTER_TYPE_CELLTYPES:
        return CLUSTER_TYPE_CELLTYPES
    else:
        raise ValueError("Unsupported cluster type: %s" % cluster_type)


def parse_clustering_key(clustering_key):
    """ Parse the output of format_clustering_key() """
    if clustering_key == CLUSTER_TYPE_GRAPHCLUST:
        return (clustering_key, 0)
    elif clustering_key.startswith(CLUSTER_TYPE_KMEANS):
        _, n_clusters, _ = clustering_key.split("_")
        return (CLUSTER_TYPE_KMEANS, int(n_clusters))
    elif clustering_key.startswith(CLUSTER_TYPE_KMEDOIDS):
        _, n_clusters, _ = clustering_key.split("_")
        return (CLUSTER_TYPE_KMEDOIDS, int(n_clusters))
    elif clustering_key == CLUSTER_TYPE_CELLTYPES:
        return (clustering_key, 0)
    else:
        raise ValueError("Unsupported clustering type for clustering key: %s" % clustering_key)


def humanify_clustering_key(clustering_key):
    """ Make a cluster_key string human-readable """
    cluster_type, cluster_param = parse_clustering_key(clustering_key)
    if cluster_type == CLUSTER_TYPE_GRAPHCLUST:
        return "Graph-based"
    elif cluster_type == CLUSTER_TYPE_KMEANS:
        return "K-means (K=%d)" % cluster_param
    elif cluster_type == CLUSTER_TYPE_KMEDOIDS:
        return "K-medoids (K=%d)" % cluster_param
    elif cluster_type == CLUSTER_TYPE_CELLTYPES:
        return "Celltypes"
    else:
        raise ValueError(
            "Unsupported clustering type %s for clustering key: %s" % (cluster_type, clustering_key)
        )


def relabel_by_size(labels):
    """Relabel clusters so they are sorted by number of members, descending.
    Args: labels (np.array(int)): 1-based cluster labels"""
    order = np.argsort(np.argsort(-np.bincount(labels)))
    return 1 + order[labels]


def save_clustering_csv(base_dir, clustering_key, labels, barcodes):
    out_dir = os.path.join(base_dir, clustering_key)
    cr_io.makedirs(out_dir, allow_existing=True)

    clusters_fn = os.path.join(out_dir, "clusters.csv")

    header = ["Barcode", "Cluster"]
    analysis_io.save_matrix_csv(clusters_fn, labels, header, barcodes)


def create_legacy_kmeans_nodes(f, new_group_name, legacy_group_name, namedtuple, clustering_key):
    """Soft-link a legacy-structured (CR 1.2) kmeans subgroup (dest) to a new-style (CR 1.3) subgroup (src).
    The old-style was a group called 'kmeans' with subgroups named _K.
    The new-style is a group called 'clustering' with subgroups named kmeans_K_clusters, etc."""
    group = f.create_group(f.root, legacy_group_name)

    cluster_type, cluster_param = parse_clustering_key(clustering_key)
    if cluster_type != CLUSTER_TYPE_KMEANS:
        return

    legacy_key = format_legacy_clustering_key(cluster_type, cluster_param)
    subgroup = f.create_group(group, legacy_key)
    for field in namedtuple._fields:
        target = "/%s/_%s/%s" % (new_group_name, clustering_key, field)

        if f.__contains__(target):
            # NOTE: coerce `target` to 'str' here because pytables chokes on unicode `target`
            f.create_soft_link(subgroup, field, target=str(target))
        else:
            sys.stderr.write(
                "Skipped soft-link of legacy dataset to %s; node doesn't exist\n" % target
            )


def subselect_barcodes(clustering, bc_indices):
    """
    Args:
        clustering (CLUSTERS namedtuple)
    """
    return CLUSTERING(
        clusters=clustering.clusters[bc_indices],
        num_clusters=clustering.num_clusters,
        cluster_score=clustering.cluster_score,
        clustering_type=clustering.clustering_type,
        global_sort_key=clustering.global_sort_key,
        description=clustering.description,
    )


def get_cluster_sizes(clustering):
    """ Returns a numpy array containing cell-counts for each cluster """
    return np.bincount(clustering.clusters)[1:]


def sort_clusterings(clusterings):
    return sorted(clusterings, key=lambda x: x.global_sort_key)


def create_clustering(
    clusters, num_clusters, cluster_score, clustering_type, global_sort_key, description
):
    """ Create a clustering namedtuple. Use numpy arrays/scalars to ensure h5 compatibility """
    return CLUSTERING(
        clusters=np.asarray(clusters),
        num_clusters=np.int64(num_clusters),
        cluster_score=np.float64(cluster_score),
        clustering_type=np.string_(clustering_type),
        global_sort_key=np.float64(global_sort_key),
        description=np.string_(description),
    )



#!/usr/bin/env python
#
# Copyright (c) 2014 10X Genomics, Inc. All rights reserved.
#

"""Values used in many other modules."""

######################################################
# DO NOT add new items to this file.
#
# - If a constant is only used from a single module, put it in that module.
# - If a constant is only used in association with a particular module, put it
#   in that module.
# - If a constant is used in a bunch of places, create a module with a more
#   specifically descriptive name to put those constants.
######################################################

# What is considered a high confidence mapped read pair
HIGH_CONF_MAPQ = 60

# Distance to mate to ensure the
MIN_MATE_OFFSET_DUP_FILTER = 20

# Sequencing settings
ILLUMINA_QUAL_OFFSET = 33
DEFAULT_HIGH_MAPQ = 60

# Demultiplex settings
# DEMULTIPLEX_DEFAULT_SAMPLE_INDEX_LENGTH = 8
DEMULTIPLEX_ACCEPTED_SAMPLE_INDEX_LENGTH = [8, 10]
""" New defaults for dual-index """
DEMULTIPLEX_BARCODE_LENGTH = 14
DEMULTIPLEX_INVALID_SAMPLE_INDEX = "X"
MAX_INDICES_TO_DEMUX = 1500
""" Maximum number of sample indices allowed """
SAMPLE_INDEX_MAX_HAMMING_DISTANCE = 1
""" Maximum separation in HD between SIs of the same kind i7/i5 """

# TAG names
PROCESSED_BARCODE_TAG = "BX"
SAMPLE_INDEX_TAG = "BC"
SAMPLE_INDEX_QUAL_TAG = "QT"
TRIM_TAG = "TR"
TRIM_QUAL_TAG = "TQ"

# Parallelization settings
PARALLEL_LOCUS_SIZE = int(4e7)

#
# Settings for metrics computation
#

# Distance to consider reads far away for far chimeras
READ_MATE_FAR_DIST = 5000

# Preflight constants
BCL_PROCESSOR_FASTQ_MODE = "BCL_PROCESSOR"
ILMN_BCL2FASTQ_FASTQ_MODE = "ILMN_BCL2FASTQ"



#!/usr/bin/env python
#
# Copyright (c) 2018 10X Genomics, Inc. All rights reserved.
#
from __future__ import absolute_import, annotations, division, print_function
from collections import OrderedDict
import copy
import pathlib
import os.path
import h5py as h5
import numpy as np
import scipy.io as sp_io
import pandas as pd

pd.set_option("compute.use_numexpr", False)
import scipy.sparse as sp_sparse
from six import ensure_binary, ensure_str, ensure_text

from bmkcc.cellranger.wrapped_tables import tables
import bmkcc.tenkit.safe_json as tk_safe_json
import bmkcc.cellranger.h5_constants as h5_constants
import bmkcc.cellranger.rna.library as rna_library
from bmkcc.cellranger.feature_ref import FeatureReference, FeatureDef, GENOME_FEATURE_TAG
import bmkcc.cellranger.utils as cr_utils
import bmkcc.cellranger.hdf5 as cr_h5
import bmkcc.cellranger.io as cr_io
import bmkcc.cellranger.sparse as cr_sparse
import bmkcc.cellranger.bisect as cr_bisect  # pylint: disable=no-name-in-module

from typing import (
    Any,
    Collection,
    Container,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)

HDF5_COMPRESSION = "gzip"
# Number of elements per chunk. Here, 1 MiB / (12 bytes)
HDF5_CHUNK_SIZE = 80000

DEFAULT_DATA_DTYPE = "int32"


# some helper functions from stats
def sum_sparse_matrix(matrix, axis: int = 0) -> np.ndarray:
    """Sum a sparse matrix along an axis."""
    axis_sum = np.asarray(matrix.sum(axis=axis))  # sum along given axis
    max_dim = np.prod(axis_sum.shape)  # get the max dimension
    return axis_sum.reshape((max_dim,))  # reshape accordingly


def top_n(array: np.ndarray, n: int) -> Tuple[int, Any]:
    """Retrieve the N largest elements and their positions in a numpy ndarray.

    Args:
       array (numpy.ndarray): Array
       n (int): Number of elements

    Returns:
       list of tuple of (int, x): Tuples are (original index, value).
    """
    indices = np.argpartition(array, -n)[-n:]
    indices = indices[np.argsort(array[indices])]
    return zip(indices, array[indices])


MATRIX_H5_FILETYPE = "matrix"
MATRIX = "matrix"
MATRIX_H5_VERSION_KEY = "version"
MATRIX_H5_VERSION = 2
SOFTWARE_H5_VERSION_KEY = "software_version"

# used to distinguish from user-defined attrs introduced in aggr
MATRIX_H5_BUILTIN_ATTRS = [
    h5_constants.H5_FILETYPE_KEY,
    MATRIX_H5_VERSION_KEY,
] + h5_constants.H5_METADATA_ATTRS


class NullAxisMatrixError(Exception):
    pass


class CountMatrixView(object):
    """Supports summing a sliced CountMatrix w/o copying the whole thing"""

    def __init__(
        self,
        matrix: CountMatrix,
        feature_indices: Optional[Collection[int]] = None,
        bc_indices: Optional[Collection[int]] = None,
    ):
        self.feature_mask = np.ones(matrix.features_dim, dtype="bool")
        self.bc_mask = np.ones(matrix.bcs_dim, dtype="bool")
        self.matrix = matrix

        if feature_indices is not None:
            self.feature_mask.fill(False)
            self.feature_mask[np.asarray(feature_indices)] = True
        if bc_indices is not None:
            self.bc_mask.fill(False)
            self.bc_mask[np.asarray(bc_indices)] = True

        self._update_feature_ref()

    @property
    def bcs_dim(self) -> int:
        return np.count_nonzero(self.bc_mask)

    @property
    def features_dim(self) -> int:
        return np.count_nonzero(self.feature_mask)

    def _copy(self) -> CountMatrixView:
        """Return a copy of this view"""
        view = CountMatrixView(self.matrix)
        view.bc_mask = np.copy(self.bc_mask)
        view.feature_mask = np.copy(self.feature_mask)
        view._update_feature_ref()
        return view

    def view(self) -> CountMatrixView:
        """Return a copy of this view"""
        return self._copy()

    def sum(self, axis: Optional[int] = None):
        """Sum across an axis."""
        return cr_sparse.sum_masked(self.matrix.m, self.feature_mask, self.bc_mask, axis=axis)

    def count_ge(self, axis: Optional[int], threshold) -> Union[int, np.ndarray]:
        """Count number of elements >= X over an axis"""
        return cr_sparse.count_ge_masked(
            self.matrix.m, self.feature_mask, self.bc_mask, threshold, axis
        )

    def select_barcodes(self, indices: Union[List[int], np.ndarray]):
        """Select a subset of barcodes (by index in the original matrix) and return the resulting view"""
        view = self._copy()
        mask = np.bincount(indices, minlength=len(view.bc_mask)).astype("bool")
        mask[mask > 1] = 1
        view.bc_mask &= mask
        return view

    def select_barcodes_by_seq(self, barcode_seqs):
        # type: (List[bytes]) -> CountMatrixView
        indices = self.matrix.bcs_to_ints(barcode_seqs)
        return self.select_barcodes(indices)

    def select_barcodes_by_gem_group(self, gem_group: int) -> CountMatrixView:
        return self.select_barcodes_by_seq(
            [bc for bc in self.matrix.bcs if gem_group == cr_utils.split_barcode_seq(bc)[1]]
        )

    def _update_feature_ref(self) -> FeatureReference:
        """Make the feature reference consistent with the feature mask"""
        indices = np.flatnonzero(self.feature_mask)
        self.feature_ref = FeatureReference(
            feature_defs=[self.matrix.feature_ref.feature_defs[i] for i in indices],
            all_tag_keys=self.matrix.feature_ref.all_tag_keys,
            target_features=self.matrix.feature_ref.target_features,
        )

    def select_features(self, indices: Union[List[int], np.ndarray]) -> CountMatrixView:
        """Select a subset of features and return the resulting view."""
        view = self._copy()
        mask = np.bincount(indices, minlength=len(view.feature_mask)).astype("bool")
        mask[mask > 1] = 1
        view.feature_mask &= mask
        view._update_feature_ref()
        return view

    def select_features_by_genome(self, genome):
        """Select the subset of gene-expression features for genes in a specific genome"""
        indices = []
        for feature in self.matrix.feature_ref.feature_defs:
            if feature.feature_type == rna_library.DEFAULT_LIBRARY_TYPE:
                if feature.tags["genome"] == genome:
                    indices.append(feature.index)
        return self.select_features(indices)

    def select_features_by_genome_and_types(self, genome, feature_types):
        """Subset the features by types and genome.

        Select the subset of gene-expression features for genes in a specific genome and matching
        one of the types listed in feature_types.
        """
        indices = []
        for feature in self.matrix.feature_ref.feature_defs:
            if feature.feature_type in feature_types:
                include_feature = True
                # Genome test only applies to GEX features
                if feature.feature_type == rna_library.DEFAULT_LIBRARY_TYPE:
                    if feature.tags["genome"] != genome:
                        include_feature = False
                if include_feature:
                    indices.append(feature.index)
        return self.select_features(indices)

    def select_features_by_types(self, feature_types):
        """Subset the features by type.

        Select the subset of gene-expression features for genes in a specific genome and matching
        one of the types listed in feature_types.
        """
        indices = []
        for feature in self.matrix.feature_ref.feature_defs:
            if feature.feature_type in feature_types:
                indices.append(feature.index)
        return self.select_features(indices)

    def select_features_by_type(self, feature_type):
        """Select the subset of features with a particular feature type (e.g. "Gene Expression")"""
        return self.select_features(self.get_feature_indices_by_type(feature_type))

    def select_features_by_ids(self, feature_ids: Iterable[bytes]) -> CountMatrixView:
        return self.select_features(self.matrix.feature_ids_to_ints(feature_ids))

    def get_feature_indices_by_type(self, feature_type: str) -> List[int]:
        """Return the list of indices of features corresponding a feature type"""
        return self.matrix.feature_ref.get_indices_for_type(feature_type)

    def get_genomes(self) -> List[str]:
        """Get a list of the distinct genomes represented by gene expression features"""
        return CountMatrix._get_genomes_from_feature_ref(self.feature_ref)

    def bcs_to_ints(self, bcs: Union[List[bytes], Set[bytes]]) -> List[int]:
        # Only works when we haven't masked barcodes.
        if np.count_nonzero(self.bc_mask) != self.matrix.bcs_dim:
            raise NotImplementedError(
                "Calling bcs_to_ints on a barcode-sliced matrix view is unimplemented"
            )
        return self.matrix.bcs_to_ints(bcs)

    def ints_to_bcs(self, bc_ints: Union[List[int], np.ndarray]) -> List[bytes]:
        if bc_ints is None or len(bc_ints) == 0:
            return []
        sliced_bc_ints = np.flatnonzero(self.bc_mask)
        orig_bc_ints = sliced_bc_ints[np.asarray(bc_ints)]
        return [self.matrix.bcs[i] for i in orig_bc_ints]

    def int_to_feature_id(self, i: int) -> bytes:
        return self.feature_ref.feature_defs[i].id

    def get_shape(self) -> Tuple[int, int]:
        """Return the shape of the sliced matrix"""
        return (np.count_nonzero(self.feature_mask), np.count_nonzero(self.bc_mask))

    def get_num_nonzero(self):
        """Return the number of nonzero entries in the sliced matrix"""
        return self.count_ge(axis=None, threshold=1)

    def get_counts_per_bc(self):
        return self.sum(axis=0)


class CountMatrix(object):
    # pylint: disable=too-many-public-methods
    def __init__(
        self,
        feature_ref: FeatureReference,
        bcs: Union[Collection[bytes], Collection[str]],
        matrix: sp_sparse.spmatrix,
    ):
        # Features (genes, CRISPR gRNAs, antibody barcodes, etc.)
        self.feature_ref = feature_ref
        self.features_dim = len(feature_ref.feature_defs)
        self.feature_ids_map = {
            f.id: f.index for f in feature_ref.feature_defs
        }  # type: Dict[bytes, int]

        # Cell barcodes
        bc_array = np.array(list(bcs), dtype="S", copy=False)
        del bcs
        bc_array.flags.writeable = False
        self.bcs = bc_array  # type: np.ndarray
        (self.bcs_dim,) = self.bcs.shape
        bcs_idx = np.argsort(self.bcs).astype(np.int32)
        bcs_idx.flags.writeable = False
        self.bcs_idx = bcs_idx  # type: np.ndarray

        self.m = matrix  # type: sp_sparse.spmatrix
        assert self.m.shape[1] == len(self.bcs), "Barcodes must be equal to cols of matrix"

    def get_shape(self):
        """Return the shape of the sliced matrix"""
        return self.m.shape

    def get_num_nonzero(self):
        """Return the number of nonzero entries in the sliced matrix"""
        return self.m.nnz

    def view(self):
        """Return a view on this matrix"""
        return CountMatrixView(self)

    @classmethod
    def empty(cls, feature_ref: FeatureReference, bcs: Collection[bytes], dtype=DEFAULT_DATA_DTYPE):
        """Create an empty matrix."""
        matrix = sp_sparse.lil_matrix((len(feature_ref.feature_defs), len(bcs)), dtype=dtype)
        return cls(feature_ref=feature_ref, bcs=bcs, matrix=matrix)

    @staticmethod
    def from_legacy_v1_h5(h5_file: h5.File) -> CountMatrix:
        """Create a CountMatrix from a legacy h5py.File (format version 1)"""

        genome_arrays = []
        gene_id_arrays = []
        gene_name_arrays = []
        bc_idx_arrays = []
        feat_idx_arrays = []
        data_arrays = []

        # Map barcode string to column index in new matrix
        barcode_map = OrderedDict()

        # Construct a genome-concatenated matrix and FeatureReference
        for genome_idx, genome in enumerate(h5_file.keys()):
            g = h5_file[genome]

            n_genes = sum(len(x) for x in gene_id_arrays)

            # Offset the row (gene) indices by the number of genes seen so far
            feat_idx_arrays.append(g["indices"][:] + n_genes)

            # Offset the col (barcode) indices by the number of nonzero elements seen so far

            # Map barcode (column) indices to a single unique barcode space
            barcodes = g["barcodes"][:]
            for bc in barcodes:
                if bc not in barcode_map:
                    barcode_map[bc] = len(barcode_map)

            remapped_col_inds = np.fromiter(
                (barcode_map[bc] for bc in barcodes), count=len(barcodes), dtype="uint64"
            )

            indptr = g["indptr"][:]
            assert len(indptr) == 1 + len(remapped_col_inds)

            if genome_idx == 0:
                # For the first set of barcodes encountered, there should
                # be no change in their new indices.
                assert np.array_equal(remapped_col_inds, np.arange(len(indptr) - 1))

            # Convert from CSC to COO by expanding the indptr array out

            nz_elems_per_bc = np.diff(indptr)
            assert len(nz_elems_per_bc) == len(g["barcodes"])

            bc_idx = np.repeat(remapped_col_inds, nz_elems_per_bc)
            assert len(bc_idx) == len(g["indices"])
            assert len(bc_idx) == len(g["data"])

            bc_idx_arrays.append(bc_idx)
            data_arrays.append(g["data"][:])

            gene_id_arrays.append(g["genes"][:])
            gene_name_arrays.append(g["gene_names"][:])
            genome_arrays.append(np.repeat(genome, len(g["genes"])))

        genomes = np.concatenate(genome_arrays)
        gene_ids = np.concatenate(gene_id_arrays)
        gene_names = np.concatenate(gene_name_arrays)

        # Construct FeatureReference
        feature_defs = []
        for (gene_id, gene_name, genome) in zip(gene_ids, gene_names, genomes):
            feature_defs.append(
                FeatureDef(
                    index=len(feature_defs),
                    id=gene_id,
                    name=gene_name,
                    feature_type=rna_library.GENE_EXPRESSION_LIBRARY_TYPE,
                    tags={GENOME_FEATURE_TAG: genome},
                )
            )
        feature_ref = FeatureReference(feature_defs, [GENOME_FEATURE_TAG])

        i = np.concatenate(feat_idx_arrays)
        j = np.concatenate(bc_idx_arrays)
        data = np.concatenate(data_arrays)

        assert isinstance(barcode_map, OrderedDict)

        matrix = sp_sparse.csc_matrix((data, (i, j)), shape=(len(gene_ids), len(barcode_map)))

        return CountMatrix(
            feature_ref, barcode_map.keys(), matrix  # pylint: disable=dict-keys-not-iterating
        )

    @staticmethod
    def from_legacy_mtx(genome_dir):
        barcodes_tsv = ensure_binary(os.path.join(genome_dir, "barcodes.tsv"))
        genes_tsv = ensure_binary(os.path.join(genome_dir, "genes.tsv"))
        matrix_mtx = ensure_binary(os.path.join(genome_dir, "matrix.mtx"))
        for filepath in [barcodes_tsv, genes_tsv, matrix_mtx]:
            if not os.path.exists(filepath):
                raise IOError("Required file not found: %s" % filepath)
        barcodes = pd.read_csv(
            barcodes_tsv.encode(),
            delimiter="\t",
            header=None,
            usecols=[0],
            dtype=bytes,
            converters={0: ensure_binary},
        ).values.squeeze()
        genes = pd.read_csv(
            genes_tsv.encode(),
            delimiter="\t",
            header=None,
            usecols=[0],
            dtype=bytes,
            converters={0: ensure_binary},
        ).values.squeeze()
        feature_defs = [
            FeatureDef(idx, gene_id, None, "Gene Expression", {})
            for (idx, gene_id) in enumerate(genes)
        ]
        feature_ref = FeatureReference(feature_defs, [])

        matrix = sp_io.mmread(matrix_mtx)
        mat = CountMatrix(feature_ref, barcodes, matrix)
        return mat

    @staticmethod
    def from_v3_mtx(genome_dir):
        barcodes_tsv = ensure_str(os.path.join(genome_dir, "barcodes.tsv.gz"))
        features_tsv = ensure_str(os.path.join(genome_dir, "features.tsv.gz"))
        matrix_mtx = ensure_str(os.path.join(genome_dir, "matrix.mtx.gz"))
        for filepath in [barcodes_tsv, features_tsv, matrix_mtx]:
            if not os.path.exists(filepath):
                raise IOError("Required file not found: {}".format(filepath))
        barcodes = pd.read_csv(
            barcodes_tsv, delimiter="\t", header=None, usecols=[0], dtype=bytes
        ).values.squeeze()
        features = pd.read_csv(features_tsv, delimiter="\t", header=None)

        feature_defs = []
        for (idx, (_, r)) in enumerate(features.iterrows()):
            fd = FeatureDef(idx, r[0], r[1], r[2], {})
            feature_defs.append(fd)

        feature_ref = FeatureReference(feature_defs, [])

        matrix = sp_io.mmread(matrix_mtx)
        mat = CountMatrix(feature_ref, barcodes, matrix)
        return mat

    @staticmethod
    def load_mtx(mtx_dir):
        legacy_fn = os.path.join(mtx_dir, "genes.tsv")
        v3_fn = os.path.join(mtx_dir, "features.tsv.gz")
        if os.path.exists(legacy_fn):
            return CountMatrix.from_legacy_mtx(mtx_dir)

        if os.path.exists(v3_fn):
            return CountMatrix.from_v3_mtx(mtx_dir)

        raise IOError("Not a valid path to a feature-barcode mtx directory: '%s'" % str(mtx_dir))

    def feature_id_to_int(self, feature_id):
        # type: (bytes) -> int
        if not isinstance(feature_id, bytes):
            raise KeyError(
                "feature_id {} must be bytes, but was {}".format(feature_id, type(feature_id))
            )
        if feature_id not in self.feature_ids_map:
            raise KeyError(
                "Specified feature ID not found in matrix: {}".format(ensure_str(feature_id))
            )
        return self.feature_ids_map[feature_id]

    def feature_ids_to_ints(self, feature_ids: Iterable[bytes]) -> List[int]:
        return sorted(self.feature_id_to_int(fid) for fid in feature_ids)

    def feature_id_to_name(self, feature_id: bytes) -> str:
        idx = self.feature_id_to_int(feature_id)
        return self.feature_ref.feature_defs[idx].name

    def int_to_feature_id(self, i: int) -> bytes:
        return self.feature_ref.feature_defs[i].id

    def int_to_feature_name(self, i: int) -> str:
        return self.feature_ref.feature_defs[i].name

    def bc_to_int(self, bc: bytes) -> int:
        """Get the integer index for a barcode.

        Args:
            bc (bytes): The barcode to search for.

        Raises:
            ValueError: `barcode` was not bytes.
            KeyError: `barcode` was not found in the set.

        Returns:
            int: the barcode index.
        """
        # Don't do a conversion for bc, because it's much more efficient to just
        # load the barcodes in the correct format in the first place.  Just
        # raise an exception so we get a nice stack trace and clear error.
        if not isinstance(bc, bytes):
            raise ValueError("barcode must be bytes, but was {}".format(type(bc)))
        j = cr_bisect.bisect_left(self.bcs_idx, bc, self.bcs)
        if j >= self.bcs_dim or self.bcs[j] != bc:
            raise KeyError("Specified barcode not found in matrix: %s" % ensure_str(bc))
        return j

    def bcs_to_ints(
        self, bcs: Union[List[bytes], Set[bytes]], return_sorted: bool = True
    ) -> List[int]:
        if isinstance(bcs, set):
            bcs = list(bcs)
        n = len(self.bcs_idx) - 1
        as_str_array = np.array(bcs, dtype=bytes)
        bcs_array = np.argsort(as_str_array)
        indices = [-1] * len(bcs)
        guess = 0
        for i, v in enumerate(bcs_array):
            bc = as_str_array[v]
            if self.bcs[guess] == bc:
                pos = guess
            else:
                pos = self.bc_to_int(bc)
                guess = min(pos + 1, n)
            if return_sorted:
                indices[i] = pos
            else:
                indices[v] = pos
        return indices

    def int_to_bc(self, j: int) -> bytes:
        return self.bcs[j]

    def ints_to_bcs(self, jj: Iterable[int]) -> List[bytes]:
        return [self.int_to_bc(j) for j in jj]

    def add(self, feature_id: bytes, bc: bytes, value=1) -> None:
        """Add a count."""
        i, j = self.feature_id_to_int(feature_id), self.bc_to_int(bc)
        self.m[i, j] += value

    def get(self, feature_id: bytes, bc: bytes) -> None:
        i, j = self.feature_id_to_int(feature_id), self.bc_to_int(bc)
        return self.m[i, j]

    def merge(self, other: CountMatrix) -> None:
        """Merge this matrix with another CountMatrix"""
        assert self.features_dim == other.features_dim
        self.m += other.m

    def sort_indices(self) -> None:
        self.tocsc()
        if not self.m.has_sorted_indices:
            self.m.sort_indices()

    def save_dense_csv(self, filename):
        """Save this matrix to a dense CSV file."""
        dense_cm = pd.DataFrame(
            self.m.toarray(),
            index=[ensure_str(f.id) for f in self.feature_ref.feature_defs],
            columns=[ensure_str(bc) for bc in self.bcs],
        )
        dense_cm.to_csv(filename, index=True, header=True)

    def save_h5_file(self, filename, extra_attrs={}, sw_version=None):
        """Save this matrix to an HDF5 file, optionally with SW version"""
        with h5.File(ensure_binary(filename), "w") as f:
            f.attrs[h5_constants.H5_FILETYPE_KEY] = MATRIX_H5_FILETYPE
            f.attrs[MATRIX_H5_VERSION_KEY] = MATRIX_H5_VERSION
            # set the software version key only if it is supplied
            if sw_version:
                f.attrs[SOFTWARE_H5_VERSION_KEY] = sw_version

            # Set optional top-level attributes
            for (k, v) in extra_attrs.items():
                cr_h5.set_hdf5_attr(f, k, v)

            group = f.create_group(MATRIX)
            self.save_h5_group(group)

    def save_h5_group(self, group: h5.Group) -> None:
        """Save this matrix to an HDF5 (h5py) group."""
        self.sort_indices()

        # Save the feature reference
        feature_ref_group = group.create_group(h5_constants.H5_FEATURE_REF_ATTR)
        self.feature_ref.to_hdf5(feature_ref_group)

        # Store barcode sequences as array of ASCII strings
        cr_h5.create_hdf5_string_dataset(
            group, h5_constants.H5_BCS_ATTR, self.bcs, compression=True
        )

        for attr, dtype in h5_constants.H5_MATRIX_ATTRS.items():
            arr = np.array(getattr(self.m, attr), dtype=dtype)
            group.create_dataset(
                attr,
                data=arr,
                chunks=(HDF5_CHUNK_SIZE,),
                maxshape=(None,),
                compression=HDF5_COMPRESSION,
                shuffle=True,
            )

    @staticmethod
    def load_dims(group: h5.Group) -> Tuple[int, int, int]:
        """Load the matrix shape from an HDF5 group"""
        (rows, cols) = group[h5_constants.H5_MATRIX_SHAPE_ATTR][:]
        entries = len(group[h5_constants.H5_MATRIX_DATA_ATTR])
        return (rows, cols, entries)

    @staticmethod
    def load_dims_from_h5(filename) -> Tuple[int, int, int]:
        """Load the matrix shape from an HDF5 file"""
        filename = ensure_binary(filename)
        h5_version = CountMatrix.get_format_version_from_h5(filename)
        if h5_version == 1:
            return CountMatrix._load_dims_from_legacy_v1_h5(filename)
        else:
            with h5.File(filename, "r") as f:
                return CountMatrix.load_dims(f[MATRIX])

    @staticmethod
    def _load_dims_from_legacy_v1_h5_handle(f: h5.File) -> Tuple[int, int, int]:
        # legacy format
        genomes = f.keys()
        num_nonzero_entries = 0
        num_gene_ids = 0
        barcodes = set()
        for genome in genomes:
            g = f[genome]
            num_nonzero_entries += len(g["data"])
            num_gene_ids += len(g["genes"])
            barcodes.update(g["barcodes"])
        return (num_gene_ids, len(barcodes), num_nonzero_entries)

    @staticmethod
    def _load_dims_from_legacy_v1_h5(filename) -> Tuple[int, int, int]:
        """Load the matrix shape from a legacy h5py.File (format version 1)"""
        with h5.File(ensure_binary(filename), "r") as f:
            return CountMatrix._load_dims_from_legacy_v1_h5_handle(f)

    @staticmethod
    def get_mem_gb_from_matrix_dim(num_barcodes: int, nonzero_entries: int) -> float:
        """ Estimate memory usage of loading a matrix. """
        matrix_mem_gb = float(nonzero_entries) / h5_constants.NUM_MATRIX_ENTRIES_PER_MEM_GB
        # We store a list and a dict of the whitelist. Based on empirical obs.
        matrix_mem_gb += float(num_barcodes) / h5_constants.NUM_MATRIX_BARCODES_PER_MEM_GB

        return h5_constants.MATRIX_MEM_GB_MULTIPLIER * np.ceil(matrix_mem_gb)

    @staticmethod
    def get_mem_gb_from_group(group: h5.Group) -> float:
        """Estimate memory usage from an HDF5 group."""
        _, num_bcs, nonzero_entries = CountMatrix.load_dims(group)
        return CountMatrix.get_mem_gb_from_matrix_dim(num_bcs, nonzero_entries)

    @staticmethod
    def get_mem_gb_crconverter_estimate_from_h5(filename) -> float:
        """Estimate the amount of memory needed for the crconverter program to process a GEX matrix.

        See the commit message for a full explanation of this process.
        """
        filename = ensure_binary(filename)
        h5_version = CountMatrix.get_format_version_from_h5(filename)
        if h5_version == 1:
            # Estimate memory usage from a legacy h5py.File (format version 1)
            _, _, nonzero_entries = CountMatrix._load_dims_from_legacy_v1_h5(filename)
        else:
            with h5.File(filename, "r") as f:
                _, _, nonzero_entries = CountMatrix.load_dims(f[MATRIX])
        return CountMatrix._get_mem_gb_crconverter_estimate_from_nnz(nonzero_entries)

    @staticmethod
    def _get_mem_gb_crconverter_estimate_from_nnz(nnz: int) -> float:
        # Empirically determined baseline + ~85 bytes per nnz
        estimate = 1.691 + 8.537e-8 * nnz
        # Arbitrary safety factor
        estimate *= 1.5
        return np.ceil(estimate)

    @staticmethod
    def get_mem_gb_from_matrix_h5(filename) -> float:
        """Estimate memory usage from an HDF5 file."""
        filename = ensure_binary(filename)
        h5_version = CountMatrix.get_format_version_from_h5(filename)
        if h5_version == 1:
            return CountMatrix._get_mem_gb_from_legacy_v1_h5(filename)
        else:
            with h5.File(filename, "r") as f:
                return CountMatrix.get_mem_gb_from_group(f[MATRIX])

    @staticmethod
    def _get_mem_gb_from_legacy_v1_h5(filename) -> float:
        """Estimate memory usage from a legacy h5py.File (format version 1)"""
        _, num_bcs, nonzero_entries = CountMatrix._load_dims_from_legacy_v1_h5(filename)
        return CountMatrix.get_mem_gb_from_matrix_dim(num_bcs, nonzero_entries)

    @classmethod
    def load(cls, group: h5.Group) -> CountMatrix:
        """Load from an HDF5 group."""
        feature_ref = CountMatrix.load_feature_ref_from_h5_group(group)
        bcs = cls.load_bcs_from_h5_group(group)

        shape = group[h5_constants.H5_MATRIX_SHAPE_ATTR][:]
        data = group[h5_constants.H5_MATRIX_DATA_ATTR][:]
        indices = group[h5_constants.H5_MATRIX_INDICES_ATTR][:]
        indptr = group[h5_constants.H5_MATRIX_INDPTR_ATTR][:]

        # Check to make sure indptr increases monotonically (to catch overflow bugs)
        assert np.all(np.diff(indptr) >= 0)

        matrix = sp_sparse.csc_matrix((data, indices, indptr), shape=shape)

        return cls(feature_ref=feature_ref, bcs=bcs, matrix=matrix)

    @staticmethod
    def load_bcs_from_h5_group(group: h5.Group) -> List[bytes]:
        """Load just the barcode sequences from an h5 group."""
        if group[h5_constants.H5_BCS_ATTR].shape is not None:
            return list(group[h5_constants.H5_BCS_ATTR][:])
        return []

    @staticmethod
    def load_bcs_from_h5(filename) -> List[bytes]:
        """Load just the barcode sequences from an HDF5 group. """
        filename = ensure_binary(filename)
        h5_version = CountMatrix.get_format_version_from_h5(filename)
        if h5_version == 1:
            return CountMatrix._load_bcs_from_legacy_v1_h5(filename)
        else:
            with h5.File(filename, "r") as f:
                return CountMatrix.load_bcs_from_h5_group(f[MATRIX])

    @staticmethod
    def _load_bcs_from_legacy_v1_h5(filename) -> List[bytes]:
        """Load just the barcode sequences from a legacy h5py.File (format version 1)"""
        with h5.File(ensure_binary(filename), "r") as f:
            genomes = f.keys()
            barcodes = set()  # type: set[bytes]
            for genome in genomes:
                group = f[genome]
                barcodes.update(group["barcodes"])
            return list(barcodes)

    @staticmethod
    def load_bcs_from_h5_file(filename) -> List[bytes]:
        with h5.File(ensure_binary(filename), "r") as f:
            if (
                h5_constants.H5_FILETYPE_KEY not in f.attrs
                or f.attrs[h5_constants.H5_FILETYPE_KEY] != MATRIX_H5_FILETYPE
            ):
                raise ValueError("HDF5 file is not a valid matrix HDF5 file.")

            if MATRIX_H5_VERSION_KEY in f.attrs:
                version = f.attrs[MATRIX_H5_VERSION_KEY]
            else:
                version = 1

            if version > MATRIX_H5_VERSION:
                raise ValueError(
                    "Matrix HDF5 file format version (%d) is a newer version that is not supported by this version of the software."
                    % version
                )
            if version < MATRIX_H5_VERSION:
                raise ValueError(
                    "Matrix HDF5 file format version (%d) is an older version that is no longer supported."
                    % version
                )

            if "matrix" not in f.keys():
                raise ValueError('Could not find the "matrix" group inside the matrix HDF5 file.')

            return CountMatrix.load_bcs_from_h5_group(f["matrix"])

    @staticmethod
    def load_library_types_from_h5_file(filename) -> Set[str]:
        """ Return a set of all library types defined in the Feature Reference """
        with h5.File(ensure_binary(filename), "r") as f:
            version = CountMatrix._get_format_version_from_handle(f)
            if version < MATRIX_H5_VERSION:
                # Only GEX supported, check for any data and return GEX if any exists
                (gene_count, _, _) = CountMatrix._load_dims_from_legacy_v1_h5_handle(f)
                if gene_count > 0:
                    return {rna_library.GENE_EXPRESSION_LIBRARY_TYPE}
                else:
                    return set()
            else:
                feature_ref = CountMatrix.load_feature_ref_from_h5_group(f[MATRIX])
                return set(f.feature_type for f in feature_ref.feature_defs)

    @staticmethod
    def load_feature_ref_from_h5_group(group: h5.Group) -> FeatureReference:
        """Load just the FeatureRef from an h5py.Group."""
        feature_group = group[h5_constants.H5_FEATURE_REF_ATTR]
        return FeatureReference.from_hdf5(feature_group)

    @staticmethod
    def load_feature_ref_from_h5_file(filename) -> FeatureReference:
        """Load just the FeatureRef from a matrix HDF5 file."""
        with h5.File(ensure_binary(filename), "r") as f:
            version = CountMatrix._get_format_version_from_handle(f)
            if version < MATRIX_H5_VERSION:
                raise IOError("Direct Feature Ref reading not supported for older H5 files.")
            else:
                return CountMatrix.load_feature_ref_from_h5_group(f[MATRIX])

    def tolil(self):
        if type(self.m) is not sp_sparse.lil_matrix:
            self.m = self.m.tolil()

    def tocoo(self):
        if type(self.m) is not sp_sparse.coo_matrix:
            self.m = self.m.tocoo()

    def tocsc(self):
        # Convert from lil to csc matrix for efficiency when analyzing data
        if type(self.m) is not sp_sparse.csc_matrix:
            self.m = self.m.tocsc()

    def select_axes_above_threshold(
        self, threshold: int = 0
    ) -> Tuple[CountMatrix, np.ndarray, np.ndarray]:
        """Select axes with sums greater than the threshold value.

        Returns:
            (CountMatrix, np.array of int, np.array of int):
                New count matrix, non-zero bc indices, feat indices
        """

        new_mat = copy.deepcopy(self)

        nonzero_bcs = np.flatnonzero(new_mat.get_counts_per_bc() > threshold)
        if self.bcs_dim > len(nonzero_bcs):
            new_mat = new_mat.select_barcodes(nonzero_bcs)

        nonzero_features = np.flatnonzero(new_mat.get_counts_per_feature() > threshold)
        if new_mat.features_dim > len(nonzero_features):
            new_mat = new_mat.select_features(nonzero_features)

        if len(nonzero_bcs) == 0 or len(nonzero_features) == 0:
            raise NullAxisMatrixError()

        return new_mat, nonzero_bcs, nonzero_features

    def select_nonzero_axes(self) -> Tuple[CountMatrix, np.ndarray, np.ndarray]:
        """Select axes with nonzero sums.

        Returns:
            (CountMatrix, np.array of int, np.array of int):
                New count matrix, non-zero bc indices, feat indices
        """
        return self.select_axes_above_threshold(0)

    def select_nonzero_axis(self, axis: int) -> Tuple[CountMatrix, np.ndarray]:
        """Select axis with nonzero sums.

        Args:
            axis (int): 0 for rows (features) and 1 for columns (barcodes).

        Returns:
            (CountMatrix, np.array of int, np.array of int):
                New count matrix, selected indices
        """

        if axis == 0:
            counts = self.get_counts_per_feature()
            indices = np.flatnonzero(counts > 0)
            return self.select_features(indices), indices

        elif axis == 1:
            counts = self.get_counts_per_bc()
            indices = np.flatnonzero(counts > 0)
            return self.select_barcodes(indices), indices

        else:
            raise ValueError("axis out of range")

    def select_barcodes(self, indices: List[int]) -> CountMatrix:
        """Select a subset of barcodes and return the resulting CountMatrix."""
        return CountMatrix(
            feature_ref=self.feature_ref,
            bcs=[self.bcs[i] for i in indices],
            matrix=self.m[:, indices],
        )

    def select_barcodes_by_seq(self, barcode_seqs: List[bytes]) -> CountMatrix:
        indices = self.bcs_to_ints(barcode_seqs, False)
        return self.select_barcodes(indices)

    def select_barcodes_by_gem_group(self, gem_group: int) -> CountMatrix:
        return self.select_barcodes_by_seq(
            [bc for bc in self.bcs if gem_group == cr_utils.get_gem_group_from_barcode(bc)]
        )

    def select_features(self, indices: Iterable[int]) -> CountMatrix:
        """Select a subset of features and return the resulting matrix.

        We also update FeatureDefs to keep their indices consistent with their new position.
        """
        feature_ref = self.feature_ref.select_features(indices)
        return CountMatrix(feature_ref=feature_ref, bcs=self.bcs, matrix=self.m[indices, :])

    def select_features_by_ids(self, feature_ids: Iterable[bytes]) -> CountMatrix:
        return self.select_features(self.feature_ids_to_ints(feature_ids))

    def remove_genes_not_on_list(self, gene_indices_to_keep: Iterable[int]) -> CountMatrix:
        """Removes all features that are GEX and not on this list, keeping all others.

        Used to subset the matrix down in the targeted assay

        Args:
            gene_indices_to_keep: list of indices of the GEX features to keep

        Returns:
            A copy of this matrix subset to the GEX feature indices requested and all other
            feature types.
        """
        indices = list(gene_indices_to_keep)
        for feature in self.feature_ref.feature_defs:
            if feature.feature_type != rna_library.GENE_EXPRESSION_LIBRARY_TYPE:
                indices.append(feature.index)
        return self.select_features(indices)

    def select_features_by_genome(self, genome: str) -> CountMatrix:
        """Select the subset of gene-expression features for genes in a specific genome"""
        indices = []
        for feature in self.feature_ref.feature_defs:
            if feature.feature_type == rna_library.DEFAULT_LIBRARY_TYPE:
                if feature.tags["genome"] == genome:
                    indices.append(feature.index)
        return self.select_features(indices)

    def select_features_by_types(self, feature_types: Container[str]) -> CountMatrix:
        """Select the subset of gene-expression features by genome and feature type."""
        indices = []
        for feature in self.feature_ref.feature_defs:
            if feature.feature_type in feature_types:
                indices.append(feature.index)
        return self.select_features(indices)

    def select_features_by_type(self, feature_type: str) -> CountMatrix:
        """Select the subset of features with a particular feature type (e.g. "Gene Expression")"""
        indices = []
        for feature in self.feature_ref.feature_defs:
            if feature.feature_type == feature_type:
                indices.append(feature.index)
        return self.select_features(indices)

    def get_feature_ids_by_type(self, feature_type: str) -> List[bytes]:
        """ Return a list of feature ids of a particular feature type (e.g. "Gene Expression")"""
        return self.feature_ref.get_feature_ids_by_type(feature_type)

    def get_count_of_feature_type(self, feature_type: str) -> int:
        """ Count how many features in the matrix are of a given type. (e.g. "Gene Expression")"""
        return self.feature_ref.get_count_of_feature_type(feature_type)

    @staticmethod
    def _get_genomes_from_feature_ref(feature_ref):
        # type: (FeatureReference) -> List[str]
        """Get a list of the distinct genomes represented by gene expression features"""
        return feature_ref.get_genomes(feature_type=rna_library.DEFAULT_LIBRARY_TYPE)

    def get_genomes(self):
        """Get a list of the distinct genomes represented by gene expression features"""
        return CountMatrix._get_genomes_from_feature_ref(self.feature_ref)

    @staticmethod
    def get_genomes_from_h5(filename) -> List[str]:
        """Get a list of the distinct genomes from a matrix HDF5 file"""
        filename = ensure_binary(filename)
        h5_version = CountMatrix.get_format_version_from_h5(filename)
        if h5_version == 1:
            return CountMatrix._get_genomes_from_legacy_v1_h5(filename)
        else:
            with h5.File(filename, "r") as f:
                feature_ref = CountMatrix.load_feature_ref_from_h5_group(f[MATRIX])
                return CountMatrix._get_genomes_from_feature_ref(feature_ref)

    @staticmethod
    def _get_genomes_from_legacy_v1_h5(filename):
        """Get a list of the distinct genomes from a legacy h5py.File (format version 1)"""
        with h5.File(ensure_binary(filename), "r") as f:
            return list(f.keys())

    def get_unique_features_per_bc(self) -> np.ndarray:
        return sum_sparse_matrix(self.m[self.m > 0], axis=0)

    def get_numfeatures_per_bc(self) -> np.ndarray:
        return sum_sparse_matrix(self.m > 0, axis=0)

    def get_counts_per_bc(self) -> np.ndarray:
        return sum_sparse_matrix(self.m, axis=0)

    def get_counts_per_barcode_for_genome(self, genome, feature_type=None) -> np.ndarray:
        """Sum the count matrix across feature rows with a given genome tag.

        The feature reference
        must contain a 'genome' tag. If feature_type is not null filter on it as well.
        """
        assert "genome" in self.feature_ref.all_tag_keys, "feature reference missing 'genome' tag"
        if feature_type:
            indices = [
                i
                for i, fdef in enumerate(self.feature_ref.feature_defs)
                if fdef.tags["genome"] == genome and fdef.feature_type == feature_type
            ]
        else:
            indices = [
                i
                for i, fdef in enumerate(self.feature_ref.feature_defs)
                if fdef.tags["genome"] == genome
            ]
        if indices:
            view = CountMatrixView(self, feature_indices=indices, bc_indices=None)
            return view.sum(axis=0)
        return np.array([], dtype=self.m.dtype)

    def get_counts_per_feature(self) -> np.ndarray:
        return sum_sparse_matrix(self.m, axis=1)

    def get_mean_and_var_per_feature(self) -> Tuple[np.ndarray, np.ndarray]:
        """Calculate the mean and variance on the sparse matrix efficiently.

        :return: a tuple with numpy arrays for mean and var
        """
        assert isinstance(self.m, sp_sparse.csc_matrix)
        mean_per_feature = self.m.mean(axis=1)
        second_moment = self.m.copy()
        second_moment = second_moment.power(2.0)
        var_per_feature = second_moment.sum(axis=1) / second_moment.shape[1] - np.power(
            mean_per_feature, 2.0
        )
        var_per_feature = np.asarray(var_per_feature)
        return (mean_per_feature, var_per_feature)

    def get_subselected_counts(
        self, list_feature_ids=None, list_barcodes=None, log_transform=False, library_type=None
    ) -> np.ndarray:
        """Get counts per barcode, sliced various ways

        - subset by list of feature IDs
        - subset by list of barcodes
        - subset by library_type
        """
        subselect_matrix = copy.deepcopy(self)

        if library_type is not None:
            assert (
                library_type in rna_library.RECOGNIZED_FEATURE_TYPES
            ), "library_type not recognized"
            subselect_matrix = subselect_matrix.select_features_by_type(library_type)

        if list_feature_ids is not None:
            subselect_matrix = subselect_matrix.select_features_by_ids(list_feature_ids)

        if list_barcodes is not None:
            subselect_matrix = subselect_matrix.select_barcodes_by_seq(list_barcodes)

        counts_feature = subselect_matrix.get_counts_per_bc()

        if log_transform:
            return np.log10(1.0 + counts_feature)
        return counts_feature

    def get_numbcs_per_feature(self) -> np.ndarray:
        return sum_sparse_matrix(self.m > 0, axis=1)

    def get_top_bcs(self, cutoff) -> np.ndarray:
        reads_per_bc = self.get_counts_per_bc()
        index = max(0, min(reads_per_bc.size, cutoff) - 1)
        value = sorted(reads_per_bc, reverse=True)[index]
        return np.nonzero(reads_per_bc >= value)[0]

    def save_mex(self, base_dir, save_features_func, metadata=None, compress=True):
        """Save in Matrix Market Exchange format.

        Note:
            This operation modifies the matrix by
            converting to a coordinate representation by calling scipy.sparse.csc_matrix.tocoo().

        Args:
          base_dir (str): Path to directory to write files in.
          save_features_func (func): Func that takes (FeatureReference, base_dir, compress) and writes
                                     a file describing the features.
          metadata (dict): Optional metadata to encode into the comments as JSON.
        """
        self.sort_indices()
        self.tocoo()

        cr_io.makedirs(base_dir, allow_existing=True)

        out_matrix_fn = os.path.join(base_dir, "matrix.mtx")
        out_barcodes_fn = os.path.join(base_dir, "barcodes.tsv")
        if compress:
            out_matrix_fn += ".gz"
            out_barcodes_fn += ".gz"

        # This method only supports an integer matrix.
        assert self.m.dtype in ["uint32", "int32", "uint64", "int64"]
        assert isinstance(self.m, sp_sparse.coo.coo_matrix)

        rows, cols = self.m.shape
        # Header fields in the file
        rep = "coordinate"
        field = "integer"
        symmetry = "general"

        metadata = metadata or {}
        metadata.update({"format_version": MATRIX_H5_VERSION})

        metadata_str = tk_safe_json.safe_jsonify(metadata)
        comment = b"metadata_json: %s" % ensure_binary(metadata_str)

        with cr_io.open_maybe_gzip(out_matrix_fn, "wb") as stream:
            # write initial header line
            stream.write(
                np.compat.asbytes(
                    "%%MatrixMarket matrix {0} {1} {2}\n".format(rep, field, symmetry)
                )
            )

            # write comments
            for line in comment.split(b"\n"):
                stream.write(b"%")
                stream.write(ensure_binary(line))
                stream.write(b"\n")

            # write shape spec
            stream.write(np.compat.asbytes("%i %i %i\n" % (rows, cols, self.m.nnz)))
            # write row, col, val in 1-based indexing
            for r, c, d in zip(self.m.row + 1, self.m.col + 1, self.m.data):
                stream.write(np.compat.asbytes(("%i %i %i\n" % (r, c, d))))

        # both GEX and ATAC provide an implementation of this in respective feature_ref.py
        save_features_func(self.feature_ref, base_dir, compress=compress)

        with cr_io.open_maybe_gzip(out_barcodes_fn, "wb") as f:
            for bc in self.bcs:
                f.write(bc + b"\n")

    @staticmethod
    def _get_format_version_from_handle(ofile):
        # type: (h5.File) -> int
        if MATRIX_H5_VERSION_KEY in ofile.attrs:
            version = ofile.attrs[MATRIX_H5_VERSION_KEY]
        else:
            version = 1
        return version

    @staticmethod
    def get_format_version_from_h5(filename):
        with h5.File(ensure_binary(filename), "r") as f:
            return CountMatrix._get_format_version_from_handle(f)

    @staticmethod
    def load_h5_file(filename):
        if isinstance(filename, pathlib.PosixPath):
            fn = filename
        else:
            fn = ensure_binary(filename)

        with h5.File(fn, "r") as f:
            if (
                h5_constants.H5_FILETYPE_KEY not in f.attrs
                or ensure_text(f.attrs[h5_constants.H5_FILETYPE_KEY]) != MATRIX_H5_FILETYPE
            ):
                raise ValueError("HDF5 file is not a valid matrix HDF5 file.")
            version = CountMatrix._get_format_version_from_handle(f)
            if version > MATRIX_H5_VERSION:
                raise ValueError(
                    "Matrix HDF5 file format version (%d) is a newer version that is not supported by this version of the software."
                    % version
                )
            if version < MATRIX_H5_VERSION:
                # raise ValueError('Matrix HDF5 file format version (%d) is an older version that is no longer supported.' % version)
                return CountMatrix.from_legacy_v1_h5(f)

            if MATRIX not in f.keys():
                raise ValueError('Could not find the "matrix" group inside the matrix HDF5 file.')

            return CountMatrix.load(f[MATRIX])

    @staticmethod
    def count_cells_from_h5(filename):
        # NOTE - this double-counts doublets.
        _, bcs, _ = CountMatrix.load_dims_from_h5(filename)
        return bcs

    @staticmethod
    def load_chemistry_from_h5(filename):
        with tables.open_file(ensure_binary(filename), "r") as f:
            try:
                chemistry = f.get_node_attr("/", h5_constants.H5_CHEMISTRY_DESC_KEY)
            except AttributeError:
                chemistry = "Unknown"
        return chemistry

    def filter_barcodes(self, bcs_per_genome):
        # type: (Dict[Any, Iterable[bytes]]) -> CountMatrix
        """Return CountMatrix containing only the specified barcodes.

        Args:
            bcs_per_genome (dict of str to list): Maps genome to cell-associated barcodes.

        Returns:
            CountMatrix w/ the specified barcodes.
        """

        # Union all the cell-associated barcodes
        bcs = set()
        for x in bcs_per_genome.values():
            bcs |= x
        bcs = list(sorted(bcs))
        return self.select_barcodes_by_seq(bcs)

    @staticmethod
    def h5_path(base_path):
        return os.path.join(base_path, "hdf5", "matrices.hdf5")


def merge_matrices(h5_filenames):
    matrix = None
    for h5_filename in h5_filenames:
        if matrix is None:
            matrix = CountMatrix.load_h5_file(h5_filename)
        else:
            other = CountMatrix.load_h5_file(h5_filename)
            matrix.merge(other)
    if matrix is not None:
        matrix.tocsc()
    return matrix


def make_matrix_attrs_count(sample_id, gem_groups, chemistry):
    matrix_attrs = make_library_map_count(sample_id, gem_groups)
    matrix_attrs[h5_constants.H5_CHEMISTRY_DESC_KEY] = chemistry
    return matrix_attrs


def load_matrix_h5_metadata(filename):
    """Get matrix metadata attributes from an HDF5 file"""
    # TODO: Consider moving these to the MATRIX key instead of the root group
    filename = ensure_binary(filename)
    h5_version = CountMatrix.get_format_version_from_h5(filename)
    if h5_version == 1:
        return _load_matrix_legacy_v1_h5_metadata(filename)
    else:
        attrs = {}
        with h5.File(filename, "r") as f:
            for key in h5_constants.H5_METADATA_ATTRS:
                val = f.attrs.get(key)
                if val is not None:
                    if np.isscalar(val) and hasattr(val, "item"):
                        # Coerce numpy scalars to python types.
                        # In particular, force np.unicode_ to unicode (python2)
                        #    or str (python 3)
                        # pylint: disable=no-member
                        attrs[key] = val.item()
                    else:
                        attrs[key] = val
            return attrs


# (needed for compatibility with v1 matrices)
def _load_matrix_legacy_v1_h5_metadata(filename):
    attrs = {}
    with tables.open_file(filename, "r") as f:
        all_attrs = f.get_node("/")._v_attrs
        for key in h5_constants.H5_METADATA_ATTRS:
            if hasattr(all_attrs, key):
                val = getattr(all_attrs, key)
                if np.isscalar(val) and hasattr(val, "item"):
                    # Coerce numpy scalars to python types.
                    # In particular, force np.unicode_ to unicode (python2)
                    #    or str (python 3)
                    attrs[key] = val.item()
                else:
                    attrs[key] = val
    return attrs


def load_matrix_h5_custom_attrs(filename):
    """Get matrix metadata attributes from an HDF5 file"""
    filename = ensure_binary(filename)
    h5_version = CountMatrix.get_format_version_from_h5(filename)
    if h5_version == 1:
        # no support for custom attrs in older versions
        return {}

    attrs = {}
    with h5.File(filename, "r") as f:
        for key, val in f.attrs.items():
            if key not in MATRIX_H5_BUILTIN_ATTRS:
                attrs[key] = val
        return attrs


def make_library_map_count(sample_id, gem_groups):
    # store gem group mapping for use by Cell Loupe
    unique_gem_groups = sorted(set(gem_groups))
    library_map = {
        h5_constants.H5_LIBRARY_ID_MAPPING_KEY: np.array(
            [sample_id] * len(unique_gem_groups), dtype="S"
        ),
        h5_constants.H5_ORIG_GEM_GROUP_MAPPING_KEY: np.array(unique_gem_groups, dtype=int),
    }
    return library_map


def make_library_map_aggr(gem_group_index) -> Dict[str, np.ndarray]:
    # store gem group mapping for use by Cell Loupe
    library_ids = []
    original_gem_groups = []
    # Sort numerically by new gem group
    for _, (lid, og) in sorted(gem_group_index.items(), key=lambda pair: int(pair[0])):
        library_ids.append(lid)
        original_gem_groups.append(og)
    library_map = {
        h5_constants.H5_LIBRARY_ID_MAPPING_KEY: np.array(library_ids, dtype="S"),
        h5_constants.H5_ORIG_GEM_GROUP_MAPPING_KEY: np.array(original_gem_groups, dtype=int),
    }
    return library_map


def get_gem_group_index(matrix_h5):
    with tables.open_file(matrix_h5, mode="r") as f:
        try:
            library_ids = f.get_node_attr("/", h5_constants.H5_LIBRARY_ID_MAPPING_KEY)
            original_gem_groups = f.get_node_attr("/", h5_constants.H5_ORIG_GEM_GROUP_MAPPING_KEY)
        except AttributeError:
            return None
    library_map = {}
    for ng, (lid, og) in enumerate(zip(library_ids, original_gem_groups), start=1):
        library_map[ng] = (lid, og)
    return library_map


def inplace_csc_column_normalize_l2(X):
    """Perform in-place column L2-normalization of input matrix X

    >>> import numpy as np
    >>> import scipy.sparse as sp
    >>> from sklearn.preprocessing import normalize
    >>> a = np.arange(12, dtype='float').reshape((3, 4))
    >>> b = sp.csc_matrix(a)
    >>> inplace_csc_column_normalize_l2(b)
    >>> np.all(normalize(a, axis=0) == b)
    True
    """
    assert X.getnnz() == 0 or isinstance(X.data[0], (np.float32, float))
    for i in range(X.shape[1]):
        s = 0.0
        for j in range(X.indptr[i], X.indptr[i + 1]):
            s += X.data[j] * X.data[j]
        if s == 0.0:
            continue
        s = np.sqrt(s)
        for j in range(X.indptr[i], X.indptr[i + 1]):
            X.data[j] /= s



#!/usr/bin/env python
#
# Copyright (c) 2017 10X Genomics, Inc. All rights reserved.
#

from __future__ import absolute_import

import os
import h5py as h5
import numpy as np
from six import ensure_binary, iteritems

import bmkcc.cellranger.analysis.io as analysis_io
import bmkcc.cellranger.analysis.constants as analysis_constants
import bmkcc.cellranger.matrix as cr_matrix

from bmkcc.cellranger.analysis.bhtsne import TSNE
import bmkcc.cellranger.analysis.clustering as cr_clustering
from bmkcc.cellranger.analysis.diffexp import DIFFERENTIAL_EXPRESSION
from bmkcc.cellranger.analysis.pca import PCA
from bmkcc.cellranger.analysis.lsa import LSA
from bmkcc.cellranger.analysis.plsa_type import PLSA
from bmkcc.cellranger.wrapped_tables import tables

COMPONENTS = {
    "pca": analysis_constants.PCA_N_COMPONENTS_DEFAULT,
    "lsa": analysis_constants.LSA_N_COMPONENTS_DEFAULT,
    "plsa": analysis_constants.PLSA_N_COMPONENTS_DEFAULT,
}


class SingleGenomeAnalysis:
    def __init__(self, matrix, method):
        self.matrix = matrix

        # parameters
        self.method = method
        self.random_state = analysis_constants.RANDOM_STATE
        self.min_n_clusters = analysis_constants.MIN_N_CLUSTERS
        self.max_n_clusters = analysis_constants.MAX_N_CLUSTERS_DEFAULT
        self.n_dimensionality_reduction_components = COMPONENTS[method]
        self.tsne_input_dims = COMPONENTS[method]
        self.n_tsne_components = analysis_constants.TSNE_N_COMPONENTS
        self.perplexity = analysis_constants.TSNE_DEFAULT_PERPLEXITY
        self.theta = analysis_constants.TSNE_THETA
        self.dr_bcs = matrix.bcs_dim
        self.dr_features = matrix.features_dim

        # DR: n_components -> DR
        self.dimensionality_reducted_matrix = {}

        # clustering_key -> CLUSTERING
        self.clusterings = {}

        # DE: clustering_key -> DIFFERENTIAL_EXPRESSION
        self.differential_expression = {}

        # t-SNE: n_tsne_components -> TSNE
        self.tsne = {}

        # Fix random seed
        np.random.seed(0)

    def is_zero_matrix(self):
        return self.matrix.bcs_dim == 0 or self.matrix.features_dim == 0

    def get_cluster_sizes(self):
        return {k: cr_clustering.get_cluster_sizes(v) for k, v in iteritems(self.clusterings)}

    def _select_bc_indices(self, cell_bc_indices):
        self.matrix = self.matrix.select_barcodes(cell_bc_indices)

        self.dr_bcs = min(self.dr_bcs, len(cell_bc_indices))
        if self.method == "pca":
            for n_components, pca in iteritems(self.dimensionality_reducted_matrix):
                self.dimensionality_reducted_matrix[n_components] = PCA(
                    pca.transformed_pca_matrix[cell_bc_indices, :],
                    pca.components,
                    pca.variance_explained,
                    pca.dispersion,
                    pca.features_selected,
                )
        elif self.method == "lsa":
            for n_components, lsa in iteritems(self.dimensionality_reducted_matrix):
                self.dimensionality_reducted_matrix[n_components] = LSA(
                    lsa.transformed_lsa_matrix[cell_bc_indices, :],
                    lsa.components,
                    lsa.variance_explained,
                    lsa.dispersion,
                    lsa.features_selected,
                )
        elif self.method == "plsa":
            for n_components, plsa in iteritems(self.dimensionality_reducted_matrix):
                self.dimensionality_reducted_matrix[n_components] = PLSA(
                    plsa.transformed_plsa_matrix[cell_bc_indices, :],
                    plsa.components,
                    plsa.variance_explained,
                    plsa.dispersion,
                    plsa.features_selected,
                )

        # Subset all the cluster label vectors
        for key, clustering in iteritems(self.clusterings):
            self.clusterings[key] = cr_clustering.subselect_barcodes(clustering, cell_bc_indices)

        for name, tsne in iteritems(self.tsne):
            self.tsne[name] = TSNE(
                tsne.transformed_tsne_matrix[cell_bc_indices, :], name=tsne.name, key=tsne.key
            )

    def subsample_bcs(self, num_bcs):
        """ Subsample barcodes across entire analysis (matrix, DR, etc) """
        if num_bcs >= self.matrix.bcs_dim:
            return

        cell_bc_indices = np.sort(
            np.random.choice(np.arange(self.matrix.bcs_dim), size=num_bcs, replace=False)
        )
        self._select_bc_indices(cell_bc_indices)

    def get_dimensionality_reduced_matrix(self, n_components=None):
        if not n_components:
            if self.method in COMPONENTS:
                n_components = COMPONENTS[self.method]
            else:
                raise ValueError("method not found")
        return self.dimensionality_reducted_matrix[n_components]

    def get_clustering(self, cluster_key):
        return self.clusterings[cluster_key]

    def get_tsne(self, key=analysis_constants.TSNE_DEFAULT_KEY):
        # Backward compatibility with older analysis HDF5 files
        # for which the key ends up being None because the 'key'
        # dataset was not generated in older versions.
        if analysis_constants.TSNE_DEFAULT_KEY not in self.tsne and None in self.tsne:
            key = None

        return self.tsne[key]

    @staticmethod
    def load_h5(filename, method):

        # CountMatrix uses h5py, not pytables
        with h5.File(ensure_binary(filename), "r") as f:
            group = f[analysis_constants.ANALYSIS_H5_MATRIX_GROUP]
            matrix = cr_matrix.CountMatrix.load(group)

        analysis = SingleGenomeAnalysis(matrix, method)

        with tables.open_file(filename, "r") as f:
            grp = None
            if method == "pca":
                grp = analysis_constants.ANALYSIS_H5_PCA_GROUP
            elif method == "lsa":
                grp = analysis_constants.ANALYSIS_H5_LSA_GROUP
            elif method == "plsa":
                grp = analysis_constants.ANALYSIS_H5_PLSA_GROUP
            else:
                raise ValueError("method invalid")
            group = f.root._v_groups[grp]
            analysis._load_dimensionality_reduced_matrix_h5(group)

            group = f.root._v_groups[analysis_constants.ANALYSIS_H5_CLUSTERING_GROUP]
            analysis._load_clustering_h5(group)

            group = f.root._v_groups[analysis_constants.ANALYSIS_H5_DIFFERENTIAL_EXPRESSION_GROUP]
            analysis._load_differential_expression_h5(group)

            group = f.root._v_groups[analysis_constants.ANALYSIS_H5_TSNE_GROUP]
            analysis._load_tsne_h5(group)

        return analysis

    def _load_dimensionality_reduced_matrix_h5(self, group):
        if self.method == "pca":
            for n_components, pca in analysis_io.load_h5_iter(group, PCA):
                self.dimensionality_reducted_matrix[int(n_components)] = pca
            return
        elif self.method == "lsa":
            for n_components, lsa in analysis_io.load_h5_iter(group, LSA):
                self.dimensionality_reducted_matrix[int(n_components)] = lsa
            return
        elif self.method == "plsa":
            for n_components, plsa in analysis_io.load_h5_iter(group, PLSA):
                self.dimensionality_reducted_matrix[int(n_components)] = plsa
            return
        raise ValueError("method {} not allowed".format(self.method))

    def _load_clustering_h5(self, group):
        for clustering_key, clustering in analysis_io.load_h5_iter(group, cr_clustering.CLUSTERING):
            self.clusterings[clustering_key] = clustering

    def _load_differential_expression_h5(self, group):
        for clustering_key, de in analysis_io.load_h5_iter(group, DIFFERENTIAL_EXPRESSION):
            self.differential_expression[clustering_key] = de

    def _load_tsne_h5(self, group):
        for _, tsne in analysis_io.load_h5_iter(group, TSNE):
            self.tsne[tsne.key] = tsne

    @staticmethod
    def load_clustering_keys_from_h5(filename):
        """ Load just the clustering keys from an analysis h5 """
        with tables.open_file(filename, "r") as f:
            group = getattr(f.root, analysis_constants.ANALYSIS_H5_CLUSTERING_GROUP)
            return [node._v_name[1:] for node in group]

    @staticmethod
    def load_clustering_from_h5(filename, clustering_key):
        """ Load a single clustering from an analysis h5 """
        with tables.open_file(filename, "r") as f:
            group = getattr(f.root, analysis_constants.ANALYSIS_H5_CLUSTERING_GROUP)
            for subgroup in group:
                if subgroup._v_name == "_" + clustering_key:
                    return analysis_io.load_h5_namedtuple(subgroup, cr_clustering.CLUSTERING)
            raise ValueError(
                "Could not find clustering key: %s in HDF5 file %s" % (clustering_key, filename)
            )

    @staticmethod
    def load_bcs_from_matrix_h5(filename):
        """ Load just the barcodes from a matrix h5 """
        with h5.File(ensure_binary(filename), "r") as f:
            # Take the first group, assuming a single-genome matrix
            # TODO: fixme when we have a single matrix group
            group_name = next(iter(f.keys()))
            return cr_matrix.CountMatrix.load_bcs_from_h5_group(f[group_name])

    @staticmethod
    def load_default_format(base_dir, method):
        h5_file_path = analysis_io.h5_path(base_dir)
        if os.path.exists(h5_file_path):
            return SingleGenomeAnalysis.load_h5(h5_file_path, method)
        else:
            return None



import os
import numpy as np
import argparse
from plotly import graph_objects as go
from bmkcc.cellranger.matrix import CountMatrix
from bmkcc.cellranger.cell_calling import find_nonambient_barcodes
from bmkcc.cellranger.io import open_maybe_gzip
from bmkcc.cellranger.webshim.data import compute_sort_order,compute_plot_segments
from bmkcc.cellranger.webshim.common import build_plot_data_dict

def parseArgs():
    parser = argparse.ArgumentParser('')
    parser.add_argument('--matrix', '-m', required=True, help='raw feature barcode matrix')
    parser.add_argument('--expect-cells', '-e', default=3000, required=False,
                        type=int, help='number of expected cells')
    parser.add_argument('--basedir', '-d', default='.', help='output direction')
    return parser.parse_args()

def save_feature(feature_ref, base_dir, compress):
    out_features = os.path.join(base_dir, "features.tsv.gz")
    with open_maybe_gzip(out_features, 'w') as f:
        tmp = ['\t'.join([i.id, i.name, i.feature_type]) for i in feature_ref.feature_defs]
        f.write('\n'.join(tmp))


def callcells(matrixDir, basedir, expect_cells):
    matrix = CountMatrix.from_v3_mtx(matrixDir) 
    matrix.m = matrix.m.astype('int64')
    matrix.tocsc()
    umis_per_bc = matrix.get_counts_per_bc()
    bc_order = np.argsort(umis_per_bc)

    thres = np.quantile(umis_per_bc[bc_order][-expect_cells:], 0.99)
    orig_cell_idx = np.flatnonzero(umis_per_bc > thres/10)
    orig_cell_bcs = matrix.bcs[orig_cell_idx]

    r = find_nonambient_barcodes(matrix, orig_cell_bcs) 
    if r is None:
        cells = orig_cell_idx
    else:
        cells = np.concatenate([r.eval_bcs[r.is_nonambient], orig_cell_idx])
    print('number of cells: {}'.format(cells.shape[0]))
    filtered_matrix = matrix.select_barcodes(cells)

    filtered_matrix.save_mex(os.path.join(basedir, 'filtered_feature_bc_matrix'),
                             save_feature, compress=True)
    return matrix, filtered_matrix

def barcode_umi_plot(counts_per_bc, raw_barcodes, cell_barcodes):
    srt_order = compute_sort_order(counts_per_bc, raw_barcodes, cell_barcodes)
    sorted_bc = raw_barcodes[srt_order]
    sorted_counts = counts_per_bc[srt_order]
    plot_segments = compute_plot_segments(sorted_bc, sorted_counts, cell_barcodes)
    plot_data = [build_plot_data_dict(i, sorted_counts) for i in plot_segments]
    layout = go.Layout(xaxis={'type':'log', 'title':'Barcodes'}, 
                       yaxis={'type':'log', 'title':'UMI counts'},
                       width=800, height=600)
    fig =  go.Figure(plot_data, layout=layout)
    return fig

def writeQC(matrix, basedir):
    Estimated_Number_of_Cells = matrix.get_shape()[1]
    Median_Genes_per_Cell = int(np.median(matrix.get_numfeatures_per_bc()))
    Total_Genes_Detected = (matrix.get_counts_per_feature()>0).sum()
    Median_UMI_Counts_per_Cell = int(np.median(matrix.get_counts_per_bc()))
    r = (f'Estimated_Number_of_Cells\t{Estimated_Number_of_Cells}\n'
         f'Median_Genes_per_Cell\t{Median_Genes_per_Cell}\n'
         f'Total_Genes_Detected\t{Total_Genes_Detected}\n'
         f'Median_UMI_Counts_per_Cell\t{Median_UMI_Counts_per_Cell}\n')
    with open(os.path.join(basedir, 'filter_qc.txt'), 'w') as f:
        f.write(r)

def cellQC(matrix):
    qc = {}
    qc['NumberofCells'] = matrix.get_shape()[1]
    qc['MedianGenes'] = int(np.median(matrix.get_numfeatures_per_bc()))
    qc['TotalGenesDetected'] = (matrix.get_counts_per_feature()>0).sum()
    qc['MedianUMICountsperCell'] = int(np.median(matrix.get_counts_per_bc()))
    return qc




if __name__=='__main__':
    args = parseArgs()
    expect_cells = args.expect_cells
    basedir = args.basedir
    matrixDir = args.matrix
    matrix, filtered_matrix = callcells(matrixDir, basedir, expect_cells)
    cell_barcodes = filtered_matrix.bcs[:]
    counts_per_bc = matrix.get_counts_per_bc()
    raw_barcodes = matrix.bcs[:]
    fig = barcode_umi_plot(counts_per_bc, raw_barcodes, cell_barcodes)
    fig.write_html(os.path.join(basedir, 'rank_barcode_umi.html'),
                   config= {'displaylogo': False})
    fig.write_image(os.path.join(basedir, 'rank_barcode_umi.png'),
                    engine='kaleido', scale=5)
    fig.write_image(os.path.join(basedir, 'rank_barcode_umi.pdf'),
                    engine='kaleido')
    writeQC(filtered_matrix, basedir)
    print('completed!\n')



#!/usr/bin/env python
#
# Copyright (c) 2020 10X Genomics, Inc. All rights reserved.
#

""" Functions for calling cell-associated barcodes """

from __future__ import absolute_import, print_function, division
from collections import namedtuple
import numpy as np
import numpy.ma as ma
from .analysis.diffexp import adjust_pvalue_bh
import bmkcc.cellranger.stats as cr_stats
import bmkcc.cellranger.sgt as cr_sgt

# Number of additional barcodes to consider after the initial cell calling
N_CANDIDATE_BARCODES = 20000

# Number of partitions (max number of barcodes to consider for ambient estimation)
# N_PARTITIONS = 90000

# Drop this top fraction of the barcodes when estimating ambient.
MAX_OCCUPIED_PARTITIONS_FRAC = 0.5

# Minimum number of UMIS per barcode to consider after the initial cell calling
MIN_UMIS = 500

# Minimum ratio of UMIs to the median (initial cell call UMI) to consider after the initial cell calling
MIN_UMI_FRAC_OF_MEDIAN = 0.01

# Maximum adjusted p-value to call a barcode as non-ambient
MAX_ADJ_PVALUE = 0.01

# Minimum number of UMIS per barcode to consider after the initial cell calling for targeted GEX
TARGETED_CC_MIN_UMIS_ADDITIONAL_CELLS = 10

# Minimum number of target gene UMIs per barcode required to be called a cell for targeted GEX
TARGETED_CC_MIN_UMIS_FROM_TARGET_GENES = 1


def estimate_profile_sgt(matrix, barcode_indices, nz_feat):
    """Estimate a gene expression profile by Simple Good Turing.
    Args:
      raw_mat (sparse matrix): Sparse matrix of all counts
      barcode_indices (np.array(int)): Barcode indices to use
      nz_feat (np.array(int)): Indices of features that are non-zero at least once
    Returns:
      profile (np.array(float)): Estimated probabilities of length len(nz_feat).
    """
    # Initial profile estimate
    prof_mat = matrix[:, barcode_indices]

    profile = np.ravel(prof_mat[nz_feat, :].sum(axis=1))
    zero_feat = np.flatnonzero(profile == 0)

    # Simple Good Turing estimate
    p_smoothed, p0 = cr_sgt.sgt_proportions(profile[np.flatnonzero(profile)])

    n0 = len(zero_feat)
    if n0 == 0:
        # Renormalize in absence of 0 class
        p_smoothed = p_smoothed / p_smoothed.sum()
        p0_i = -1.0  # unused
    else:
        # Distribute p0 equally among the zero elements.
        p0_i = p0 / n0

    profile_p = np.repeat(p0_i, len(nz_feat))
    profile_p[np.flatnonzero(profile)] = p_smoothed

    assert np.isclose(profile_p.sum(), 1.0)
    return profile_p


# Construct a background expression profile from barcodes with <= T UMIs
def est_background_profile_sgt(matrix, use_bcs):
    """Estimate a gene expression profile on a given subset of barcodes.
         Use Good-Turing to smooth the estimated profile.
    Args:
      matrix (scipy.sparse.csc_matrix): Sparse matrix of all counts
      use_bcs (np.array(int)): Indices of barcodes to use (col indices into matrix)
    Returns:
      profile (use_features, np.array(float)): Estimated probabilities of length use_features.
    """
    # Use features that are nonzero anywhere in the data
    use_feats = np.flatnonzero(np.asarray(matrix.sum(1)))

    # Estimate background profile
    bg_profile_p = estimate_profile_sgt(matrix, use_bcs, use_feats)

    return (use_feats, bg_profile_p)


NonAmbientBarcodeResult = namedtuple(
    "NonAmbientBarcodeResult",
    [
        "eval_bcs",  # Candidate barcode indices (n)
        "log_likelihood",  # Ambient log likelihoods (n)
        "pvalues",  # pvalues (n)
        "pvalues_adj",  # B-H adjusted pvalues (n)
        "is_nonambient",  # Boolean nonambient calls (n)
    ],
)


def find_nonambient_barcodes(
    matrix,
    orig_cell_bcs,
    chemistry_description=None,
    min_umi_frac_of_median=MIN_UMI_FRAC_OF_MEDIAN,
    min_umis_nonambient=MIN_UMIS,
    max_adj_pvalue=MAX_ADJ_PVALUE,
):
    """Call barcodes as being sufficiently distinct from the ambient profile

    Args:
      matrix (CountMatrix): Full expression matrix.
      orig_cell_bcs (iterable of str): Strings of initially-called cell barcodes.
      chemistry_description: Change ambient RNA estimation for LT chemistry
    Returns:
    TBD
    """
    # Estimate an ambient RNA profile
    umis_per_bc = matrix.get_counts_per_bc()
    bc_order = np.argsort(umis_per_bc)

    ### TODO: The assert below is temporary while we are confirming
    ### chemistry detection works as intended. Remove before CR6.0 is shipped
    ### https://10xtech.atlassian.net/browse/CELLRANGER-4198,4199
    sorted_umis_per_bc = sorted(umis_per_bc, reverse=True)
    if chemistry_description == "Single Cell 3' v3 LT":
        assert (
            np.sum(sorted_umis_per_bc[45000:90000]) < 10
        ), "Sum of UMIs for barcodes between ranks 45000-90000 is non-negligible, yet LT chemistry was detected"
        N_PARTITIONS = 9000
    else:
        N_PARTITIONS = 90000

    # Take what we expect to be the barcodes associated w/ empty partitions.
    empty_bcs = bc_order[::-1][(N_PARTITIONS // 2) : N_PARTITIONS]
    empty_bcs.sort()

    # Require non-zero barcodes
    nz_bcs = np.flatnonzero(umis_per_bc)
    nz_bcs.sort()

    use_bcs = np.intersect1d(empty_bcs, nz_bcs, assume_unique=True)

    if len(use_bcs) > 0:
        try:
            eval_features, ambient_profile_p = est_background_profile_sgt(matrix.m, use_bcs)
        except cr_sgt.SimpleGoodTuringError as e:
            print(str(e))
            return None
    else:
        eval_features = np.zeros(0, dtype=int)
        ambient_profile_p = np.zeros(0)

    # Choose candidate cell barcodes
    orig_cell_bc_set = set(orig_cell_bcs)
    orig_cells = np.flatnonzero(
        np.fromiter(
            (bc in orig_cell_bc_set for bc in matrix.bcs), count=len(matrix.bcs), dtype=bool
        )
    )

    # No good incoming cell calls
    if orig_cells.sum() == 0:
        return None

    # Look at non-cell barcodes above a minimum UMI count
    eval_bcs = np.ma.array(np.arange(matrix.bcs_dim))
    eval_bcs[orig_cells] = ma.masked

    median_initial_umis = np.median(umis_per_bc[orig_cells])
    min_umis = int(max(min_umis_nonambient, np.ceil(median_initial_umis * min_umi_frac_of_median)))
    print("Median UMIs of initial cell calls: {}".format(median_initial_umis))
    print("Min UMIs: {}".format(min_umis))

    eval_bcs[umis_per_bc < min_umis] = ma.masked
    n_unmasked_bcs = len(eval_bcs) - eval_bcs.mask.sum()

    # Take the top N_CANDIDATE_BARCODES by UMI count, of barcodes that pass the above criteria
    eval_bcs = np.argsort(ma.masked_array(umis_per_bc, mask=eval_bcs.mask))[0:n_unmasked_bcs][
        -N_CANDIDATE_BARCODES:
    ]
    # SORT the barcodes. This is a critical step; eval_bcs is a list of integers, and these indices are used
    # to get the counts via matrix.select_features_by_genome(genome).select_barcodes(eval_bcs).get_counts_per_bc(),
    # which sorts the input, and also to get the string barcode sequences via np.array(matrix.bcs)[eval_bcs],
    # which doesn't. Without sorting here, the matching between the sequences and their counts from both
    # genomes is essentially random, thus the species assignments will be wrong.
    eval_bcs.sort()

    if len(eval_bcs) == 0:
        return None

    assert not np.any(np.isin(eval_bcs, orig_cells))
    print("Number of candidate bcs: {}".format(len(eval_bcs)))
    print(
        "Range candidate bc umis: {}, {}".format(
            umis_per_bc[eval_bcs].min(), umis_per_bc[eval_bcs].max()
        )
    )

    eval_mat = matrix.m[eval_features, :][:, eval_bcs]

    if len(ambient_profile_p) == 0:
        obs_loglk = np.repeat(np.nan, len(eval_bcs))
        pvalues = np.repeat(1, len(eval_bcs))
        sim_loglk = np.repeat(np.nan, len(eval_bcs))
        return None

    # Compute observed log-likelihood of barcodes being generated from ambient RNA
    obs_loglk = cr_stats.eval_multinomial_loglikelihoods(eval_mat, ambient_profile_p)

    # Simulate log likelihoods
    distinct_ns, sim_loglk = cr_stats.simulate_multinomial_loglikelihoods(
        ambient_profile_p, umis_per_bc[eval_bcs], num_sims=10000, verbose=True
    )

    # Compute p-values
    pvalues = cr_stats.compute_ambient_pvalues(
        umis_per_bc[eval_bcs], obs_loglk, distinct_ns, sim_loglk
    )

    pvalues_adj = adjust_pvalue_bh(pvalues)

    is_nonambient = pvalues_adj <= max_adj_pvalue

    return NonAmbientBarcodeResult(
        eval_bcs=eval_bcs,
        log_likelihood=obs_loglk,
        pvalues=pvalues,
        pvalues_adj=pvalues_adj,
        is_nonambient=is_nonambient,
    )



#!/usr/bin/env python
#
# Copyright (c) 2014 10X Genomics, Inc. All rights reserved.
#

""" Utils for math operations """

from __future__ import print_function, division, absolute_import

try:
    from typing import Iterable, Union  # pylint: disable=unused-import
except ImportError:
    pass  # python 2


def NX(lengths, fraction):
    # type: (Iterable[int], float) -> int
    """Calculates the N50 for a given set of lengths"""
    lengths = sorted(lengths, reverse=True)
    tot_length = sum(lengths)
    cum_length = 0
    for length in lengths:
        cum_length += length
        if cum_length >= fraction * tot_length:
            return length
    raise ArithmeticError("this should be impossible")


def robust_divide(a, b):
    # type: (Union[float, int, str, bytes], Union[float, int, str, bytes]) -> float
    """Handles 0 division and conversion to floats automatically"""
    a = float(a)
    b = float(b)
    if b == 0:
        return float("NaN")
    else:
        return a / b



#!/usr/bin/env python
#
# Copyright (c) 2014 10X Genomics, Inc. All rights reserved.
#

"""Methods for safely encoding values to json.

The json standard does not permit encoding NaN, but Python will still happily
do it, which can cause problems for other programs.  This module contains code
to fix those values up, as well as convert numpy objects to things which encode
properly in json.
"""

from __future__ import absolute_import

import json
import math
import numpy as np
import six

NAN_STRING = "NaN"
POS_INF_STRING = "inf"
NEG_INF_STRING = "-inf"
_NEG_INF = float("-Inf")
_POS_INF = float("+Inf")


def desanitize_value(x):
    """
    Converts back a special numeric value encoded by json_sanitize
    :param x: a string encoded value or float
    :return: a float or the original value
    """
    if x == NAN_STRING:
        return float("NaN")
    elif x == POS_INF_STRING:
        return _POS_INF
    elif x == NEG_INF_STRING:
        return _NEG_INF
    else:
        return x


def json_sanitize(data):
    """
    Yuck yuck yuck yuck yuck!
    The default JSON encoders do bad things if you try to encode
    NaN/Infinity/-Infinity as JSON.  This code takes a nested data structure
    composed of: atoms, dicts, or array-like iteratables and finds all of the
    not-really-a-number floats and converts them to an appropriate string for
    subsequent jsonification.
    """
    # This really doesn't make me happy. How many cases we we have to test?
    if isinstance(data, str):
        return data
    elif isinstance(data, bytes):
        return six.ensure_str(data)
    elif isinstance(data, (float, np.float64)):
        # Handle floats specially
        if math.isnan(data):
            return NAN_STRING
        if data == float("+Inf"):
            return POS_INF_STRING
        if data == float("-Inf"):
            return NEG_INF_STRING
        return data
    elif isinstance(data, np.integer):
        return int(data)
    elif isinstance(data, dict):
        return {json_sanitize(k): json_sanitize(value) for k, value in data.items()}
    elif hasattr(data, "keys"):
        # Dictionary-like case
        return {json_sanitize(k): json_sanitize(data[k]) for k in data.keys()}
    elif hasattr(data, "__iter__"):
        # Anything else that looks like a list. N
        return [json_sanitize(item) for item in data]
    elif hasattr(data, "shape") and data.shape == ():
        # Numpy 0-d array
        return np.asscalar(data)
    else:
        return data


def safe_jsonify(data, pretty=False, **kwargs):
    """Dump an object to a string as json, after sanitizing it."""
    safe_data = json_sanitize(data)
    if pretty:
        return json.dumps(
            safe_data,
            indent=kwargs.pop("indent", 4),
            sort_keys=kwargs.pop("sort_keys", True),
            separators=kwargs.pop("separators", (",", ":")),
            **kwargs,
        )
    else:
        return json.dumps(safe_data, **kwargs)


class NumpyAwareJSONEncoder(json.JSONEncoder):
    """This encoder will convert 1D np.ndarrays to lists.  For other numpy
    types, uses obj.item() to extract a python scalar."""

    def default(self, o):  # pylint: disable=method-hidden
        """Convert a 1D np.ndarray into a list, or a single-element array or
        matrix into its scalar value."""
        if isinstance(o, np.ndarray) and o.ndim == 1:
            return o.tolist()
        elif isinstance(o, np.integer):
            return int(o)
        elif isinstance(o, np.generic):
            return o.item()
        elif isinstance(o, bytes):
            return six.ensure_str(o)
        return json.JSONEncoder.default(self, o)


def dump_numpy(data, fp, pretty=False, **kwargs):  # pylint: disable=invalid-name
    """ Dump object to json, converting numpy objects to reasonable JSON """
    if pretty:
        json.dump(
            data,
            fp,
            cls=NumpyAwareJSONEncoder,
            indent=kwargs.pop("indent", 4),
            sort_keys=kwargs.pop("sort_keys", True),
            separators=kwargs.pop("separators", (",", ":")),
            **kwargs,
        )
    else:
        json.dump(data, fp, cls=NumpyAwareJSONEncoder, **kwargs)



#!/usr/bin/env python
#
# Copyright (c) 2017 10X Genomics, Inc. All rights reserved.
#
""" Differential expression analysis for single-cell RNA-seq """
# pylint: disable=invalid-name

from __future__ import absolute_import, division, print_function
import collections
import sys
import os
import numpy as np
import pandas as pd
from scipy.special import gammaln, logsumexp
import scipy.stats
from sklearn.utils import sparsefuncs
from six.moves import xrange as range
import bmkcc.cellranger.analysis.io as analysis_io
import bmkcc.cellranger.analysis.clustering as cr_clustering
import bmkcc.cellranger.analysis.constants as analysis_constants
import bmkcc.cellranger.io as cr_io

SSEQ_ZETA_QUANTILE = 0.995

# pylint: disable=invalid-name
DIFFERENTIAL_EXPRESSION = collections.namedtuple("DIFFERENTIAL_EXPRESSION", ["data"])


def estimate_size_factors(x):
    """Estimate size factors (related to cell RNA content and GEM-to-GEM technical variance)
    Args:
      x - Sparse matrix (csc) of counts (feature x cell)
    Returns:
      Array of floats, one per cell.
    """
    counts_per_cell = np.squeeze(np.asarray(x.sum(axis=0)))
    size_factors = counts_per_cell.astype(np.float64) / np.median(counts_per_cell)
    return size_factors


def compute_sseq_params(x, zeta_quantile=SSEQ_ZETA_QUANTILE):
    """Compute global parameters for the sSeq differential expression method.
    The key parameters are the shrunken feature-wise dispersions.

    This method was published in:
    Yu D, et al. (2013) Shrinkage estimation of dispersion in Negative Binomial models for RNA-seq experiments with small sample size.
    Bioinformatics. 29: 1275-1282. doi: 10.1093/bioinformatics/btt143
    Args:
      x - Sparse matrix (csc) of counts (feature x cell)
      zeta_quantile (float) - Quantile of method-of-moments dispersion estimates to
                              use as the shrinkage target zeta.
    Returns:
      A dictionary containing the sSeq parameters and some diagnostic info.
    """
    # Number of cells
    N = x.shape[1]

    # Number of features
    G = x.shape[0]

    # Estimate size factors and normalize the matrix for quick mean/var calcs
    size_factors = estimate_size_factors(x)
    # Cast to float to prevent truncation of 1 -> 0 for size factors < 1
    x_norm = scipy.sparse.csc_matrix(x, dtype=np.float64, copy=True)
    sparsefuncs.inplace_column_scale(x_norm, 1.0 / size_factors)

    # Estimate featurewise mean, variance, and dispersion by the method of moments
    # assuming that each feature follows a negative-binomial distribution.
    mean_g = np.squeeze(np.asarray(x_norm.mean(axis=1, dtype=np.float64)))
    # V[X] = E[X^2] - E[X]^2
    mean_sq_g = np.squeeze(np.asarray(x_norm.multiply(x_norm).mean(axis=1, dtype=np.float64)))
    var_g = mean_sq_g - np.square(mean_g)

    # Method of moments estimate of feature-wise dispersion (phi)
    # Only use features with non-zero variance in the following estimation
    use_g = var_g > 0
    phi_mm_g = np.zeros(G)
    phi_mm_g[use_g] = np.maximum(
        0,
        (float(N) * var_g[use_g] - mean_g[use_g] * np.sum(1.0 / size_factors))
        / (np.square(mean_g[use_g]) * np.sum(1.0 / size_factors)),
    )

    # Estimate the optimal global target dispersion (zeta_hat).
    # The true optimal zeta is that which minimizes the MSE vs the true dispersions.
    # The featurewise dispersions will be "shrunk" towards our estimate of zeta.

    # Use a high quantile of the MoM dispersion as our shrinkage target
    # per the rule of thumb in Yu, et al.
    zeta_hat = np.nanpercentile(phi_mm_g[use_g], 100.0 * zeta_quantile)

    # Compute delta, the optimal shrinkage towards zeta_hat
    # This defines a linear function that shrinks the MoM dispersion estimates
    mean_phi_mm_g = np.mean(phi_mm_g[use_g])
    delta = (np.sum(np.square(phi_mm_g[use_g] - mean_phi_mm_g)) / float(G - 1)) / (
        np.sum(np.square(phi_mm_g[use_g] - zeta_hat)) / float(G - 2)
    )

    # Compute the shrunken dispersion estimates
    # Interpolate between the MoM estimates and zeta_hat by delta
    phi_g = np.full(G, np.nan)
    if np.any(phi_mm_g[use_g] > 0):
        phi_g[use_g] = (1 - delta) * phi_mm_g[use_g] + delta * zeta_hat
    else:
        phi_g[use_g] = 0.0

    return {
        "N": N,
        "G": G,
        "size_factors": size_factors,
        "mean_g": mean_g,
        "var_g": var_g,
        "use_g": use_g,
        "phi_mm_g": phi_mm_g,
        "eval_zeta": None,
        "eval_asd": None,
        "asd_slope": None,
        "zeta_hat": zeta_hat,
        "delta": delta,
        "phi_g": phi_g,
    }


def get_local_sseq_params(x, group_a, group_b):
    """Compute locally-distinguishing sseq parameters.
    For perturbation vs control analysis for CRISPR
    and meta sample comparison within an aggregated matrix"""

    print("...Computing params for this comparison...")
    sys.stdout.flush()
    both_conditions = np.concatenate([group_a, group_b])
    matrix_groups = x[:, both_conditions]
    new_group = list(range(len(both_conditions)))
    new_group_a = new_group[0 : len(group_a)]
    new_group_b = new_group[len(group_a) : len(both_conditions)]
    return compute_sseq_params(matrix_groups), new_group_a, new_group_b, matrix_groups


def neg_bin_log_pmf(k, mu, phi):
    """Log(PMF) of negative binomial distribution with mean mu and dispersion phi,
    conveniently parameterized.
    Args:
      k (int) - NB random variable
      mu (float) - mean
      phi (float) - dispersion
    Returns:
      The log of the pmf at k."""
    r = 1.0 / phi
    return (
        gammaln(r + k)
        - (gammaln(r) + gammaln(k + 1))
        + k * np.log(mu / (r + mu))
        + r * np.log(r / (r + mu))
    )


def nb_exact_test(x_a, x_b, size_factor_a, size_factor_b, mu, phi):
    """Compute p-value for a pairwise exact test using the negative binomial.
    Args:
      x_a (int) - Total count for a single feature in group A
      x_b (int) - Total count for a single feature in group B
      size_factor_a (float) - Sum of size factors for group A
      size_factor_b (float) - Sum of size factors for group B
      mu (float) - Common mean count for this feature
      phi (float) - Common dispersion for this feature
    Returns:
      p-value (float); the probability that a random pair of counts under the null hypothesis is more extreme
      than the observed pair of counts."""
    size_factor_a = float(size_factor_a)
    size_factor_b = float(size_factor_b)
    mu = float(mu)
    phi = float(phi)

    if (x_a + x_b) == 0:
        return 1.0
    if phi == 0:
        return 1.0
    if size_factor_a == 0 or size_factor_b == 0:
        return 1.0

    def log_prob_all(x, sa, sb):
        """ TODO """

        all_x = np.arange(0, x + 1, 1)
        r = 1.0 / phi
        gln = gammaln(all_x + 1)
        total = gammaln(sa * r + all_x) + gammaln(sb * r + all_x[::-1]) - (gln + gln[::-1])
        total += (
            x * np.log(mu / (r + mu))
            + (sa + sb) * np.log(r / (r + mu))
            - gammaln(sa * r)
            - gammaln(sb * r)
        )
        return total

    log_p_all = log_prob_all(x_a + x_b, size_factor_a, size_factor_b)
    log_p_obs = log_p_all[x_a]

    more_extreme = log_p_all <= log_p_obs
    if np.sum(more_extreme) == 0:
        return 0.0

    return np.exp(logsumexp(log_p_all[more_extreme]) - logsumexp(log_p_all))


def nb_asymptotic_test(x_a, x_b, size_factor_a, size_factor_b, mu, phi):
    """Compute p-value for a pairwise exact test using a fast beta approximation
    to the conditional joint distribution of (x_a, x_b).
    Robinson MD and Smyth GK (2008). Small-sample estimation of negative binomial dispersion, with applications to SAGE data. Biostatistics, 9, 321-332
    "It is based a method-of-moments gamma approximation to the negative binomial distribution."
      - Personal communication w/ author
    Adapted from implementation in the "edgeR" package:
      https://github.com/Bioconductor-mirror/edgeR/blob/1ab290c9585335cf99bb41f50cfce2ce4d40f907/R/exactTestBetaApprox.R
    This function is vectorized. It always returns a vector even if the inputs are scalar.
    Args:
      x_a (int/np.array) - Total count for a single feature in group A
      x_b (int/np.array) - Total count for a single feature in group B
      size_factor_a (float) - Sum of size factors for group A
      size_factor_b (float) - Sum of size factors for group B
      mu (float/np.array) - Common mean count for this feature
      phi (float/np.array) - Common dispersion for this feature
    Returns:
      p-value (np.array); the probability that a random pair of counts under the null hypothesis is more extreme
      than the observed pair of counts."""
    x_a = np.array(x_a, ndmin=1, copy=False)
    x_b = np.array(x_b, ndmin=1, copy=False)
    mu = np.array(mu, ndmin=1, copy=False)
    phi = np.array(phi, ndmin=1, copy=False)

    alpha = size_factor_a * mu / (1.0 + phi * mu)
    beta = (size_factor_b / size_factor_a) * alpha

    total = x_a + x_b

    median = scipy.stats.beta.median(alpha, beta)
    left = ((x_a + 0.5) / total) < median
    right = np.logical_not(left)

    p = np.empty(len(x_a))
    p[left] = 2 * scipy.stats.beta.cdf((x_a[left] + 0.5) / total[left], alpha[left], beta[left])
    # If X ~ Beta(a, b) then 1 - X ~ Beta(b, a)
    # This avoids the asymmetry in beta.cdf
    p[right] = 2 * (
        scipy.stats.beta.cdf((x_b[right] + 0.5) / total[right], beta[right], alpha[right])
    )
    return p


def adjust_pvalue_bh(p):
    """ Multiple testing correction of p-values using the Benjamini-Hochberg procedure """
    descending = np.argsort(p)[::-1]
    # q = p * N / k where p = p-value, N = # tests, k = p-value rank
    scale = float(len(p)) / np.arange(len(p), 0, -1)
    # pylint: disable=no-member
    q = np.minimum(1, np.minimum.accumulate(scale * p[descending]))

    # Return to original order
    return q[np.argsort(descending)]


def numpy_logical_and_list(list_of_logicals):
    """Takes the item-wise logical-AND of a list
    of logical vectors.
    """
    assert len(list_of_logicals) >= 2
    output = list_of_logicals[0]
    for i in range(1, len(list_of_logicals)):
        output = np.logical_and(output, list_of_logicals[i])
    return output


def sseq_differential_expression(x, cond_a, cond_b, sseq_params, big_count=900):
    """Run sSeq pairwise differential expression test.
    Args:
      x - Sparse matrix (csc) of counts (feature x cell)
      cond_a (np.array(int)): Indices of cells in group A
      cond_b (np.array(int)): Indices of cells in group B
      sseq_params (dict): Precomputed global parameters
      big_count (int): Use asymptotic approximation if both counts > this
    Returns:
      A pd.DataFrame with DE results for group A relative to group B"""
    x_a = x[:, cond_a]
    x_b = x[:, cond_b]

    # Number of features
    G = x.shape[0]

    # Size factors
    size_factor_a = np.sum(sseq_params["size_factors"][cond_a])
    size_factor_b = np.sum(sseq_params["size_factors"][cond_b])

    # Compute p-value for each feature
    p_values = np.ones(G)

    feature_sums_a = np.squeeze(np.asarray(x_a.sum(axis=1)))
    feature_sums_b = np.squeeze(np.asarray(x_b.sum(axis=1)))

    big = numpy_logical_and_list(
        [sseq_params["use_g"], feature_sums_a > big_count, feature_sums_b > big_count]
    )
    small = np.logical_and(sseq_params["use_g"], np.logical_not(big))

    sys.stderr.write(
        "Computing {} exact tests and {} asymptotic tests.\n".format(np.sum(small), np.sum(big))
    )

    # Compute exact test for small-count features
    for i in np.flatnonzero(small):
        p_values[i] = nb_exact_test(
            feature_sums_a[i],
            feature_sums_b[i],
            size_factor_a,
            size_factor_b,
            sseq_params["mean_g"][i],
            sseq_params["phi_g"][i],
        )
    # Compute asymptotic approximation for big-count features
    p_values[big] = nb_asymptotic_test(
        feature_sums_a[big],
        feature_sums_b[big],
        size_factor_a,
        size_factor_b,
        sseq_params["mean_g"][big],
        sseq_params["phi_g"][big],
    )
    # Adjust p-values for multiple testing correction
    # Only adjust the features that were actually tested
    adj_p_values = p_values.copy()
    adj_p_values[sseq_params["use_g"]] = adjust_pvalue_bh(p_values[sseq_params["use_g"]])

    de_result = pd.DataFrame(
        {
            "tested": sseq_params["use_g"],
            "sum_a": feature_sums_a,
            "sum_b": feature_sums_b,
            "common_mean": sseq_params["mean_g"],
            "common_dispersion": sseq_params["phi_g"],
            "norm_mean_a": feature_sums_a / size_factor_a,
            "norm_mean_b": feature_sums_b / size_factor_b,
            "p_value": p_values,
            "adjusted_p_value": adj_p_values,
            # Introduce a pseudocount into log2(fold_change)
            "log2_fold_change": np.log2((1 + feature_sums_a) / (1 + size_factor_a))
            - np.log2((1 + feature_sums_b) / (1 + size_factor_b)),
        }
    )

    return de_result


def run_differential_expression(matrix, clusters, sseq_params=None):
    """Compute differential expression for each cluster vs all other cells
    Args: matrix      - GeneBCMatrix  :  feature expression data
          clusters    - np.array(int) :  1-based cluster labels
          sseq_params - dict          :  params from compute_sseq_params"""

    n_clusters = np.max(clusters)

    if sseq_params is None:
        print("Computing params...")
        sys.stdout.flush()
        sseq_params = compute_sseq_params(matrix.m)

    # Create a numpy array with 3*K columns;
    # each group of 3 columns is mean, log2, pvalue for cluster i
    all_de_results = np.zeros((matrix.features_dim, 3 * n_clusters))

    for cluster in range(1, 1 + n_clusters):
        in_cluster = clusters == cluster
        group_a = np.flatnonzero(in_cluster)
        group_b = np.flatnonzero(np.logical_not(in_cluster))
        print("Computing DE for cluster %d..." % cluster)
        sys.stdout.flush()

        de_result = sseq_differential_expression(matrix.m, group_a, group_b, sseq_params)
        all_de_results[:, 0 + 3 * (cluster - 1)] = de_result["norm_mean_a"]
        all_de_results[:, 1 + 3 * (cluster - 1)] = de_result["log2_fold_change"]
        all_de_results[:, 2 + 3 * (cluster - 1)] = de_result["adjusted_p_value"]

    return DIFFERENTIAL_EXPRESSION(all_de_results)


def save_differential_expression_csv(
    clustering_key, de, matrix, base_dir, cluster_names=None, file_name="differential_expression"
):
    """ Write diffexp results to CSV """

    out_dir = base_dir
    if clustering_key is not None:
        out_dir = os.path.join(base_dir, clustering_key)
    cr_io.makedirs(out_dir, allow_existing=True)

    diff_expression_fn = os.path.join(out_dir, file_name + ".csv")
    diff_expression_header = ["Feature ID", "Feature Name"]

    n_clusters = de.data.shape[1] // 3
    for i in range(n_clusters):
        if cluster_names is None:
            diff_expression_header += [
                "Cluster %d Mean Counts" % (i + 1),
                "Cluster %d Log2 fold change" % (i + 1),
                "Cluster %d Adjusted p value" % (i + 1),
            ]
        else:
            diff_expression_header += [
                "Perturbation %s, Mean Counts" % cluster_names[i],
                "Perturbation %s, Log2 fold change" % cluster_names[i],
                "Perturbation %s, Adjusted p value" % cluster_names[i],
            ]

    diff_expression_prefixes = [(f.id, f.name) for f in matrix.feature_ref.feature_defs]
    analysis_io.save_matrix_csv(
        diff_expression_fn, de.data, diff_expression_header, diff_expression_prefixes
    )


def save_differential_expression_csv_from_features(
    clustering_key, de, diff_expression_prefixes, base_dir
):
    """ Write diffexp results to CSV """
    out_dir = os.path.join(base_dir, clustering_key)
    cr_io.makedirs(out_dir, allow_existing=True)

    diff_expression_fn = os.path.join(out_dir, "differential_expression.csv")
    diff_expression_header = ["Feature ID", "Feature Name"]

    n_clusters = de.data.shape[1] // 3
    for i in range(n_clusters):
        diff_expression_header += [
            "Cluster %d Mean Counts" % (i + 1),
            "Cluster %d Log2 fold change" % (i + 1),
            "Cluster %d Adjusted p value" % (i + 1),
        ]

    analysis_io.save_matrix_csv(
        diff_expression_fn, de.data, diff_expression_header, diff_expression_prefixes
    )


def save_differential_expression_h5(f, clustering_key, de):
    """ Write diffexp results to H5File `f` """
    group = f.create_group(f.root, analysis_constants.ANALYSIS_H5_DIFFERENTIAL_EXPRESSION_GROUP)

    analysis_io.save_h5(f, group, clustering_key, de)

    cr_clustering.create_legacy_kmeans_nodes(
        f,
        analysis_constants.ANALYSIS_H5_DIFFERENTIAL_EXPRESSION_GROUP,
        analysis_constants.ANALYSIS_H5_KMEANS_DIFFERENTIAL_EXPRESSION_GROUP,
        DIFFERENTIAL_EXPRESSION,
        clustering_key,
    )
